# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
License Plate detection model with model types: yolov4 and yolov4tiny
"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.yolov4_license_plate.yolo_license_plate_files.detector import (  # pylint: disable=line-too-long
    Detector,
)


class YOLOLicensePlateModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """Yolo model with model types: v4 and v4tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds(["iou_threshold", "score_threshold"], "[0, 1]")

        model_dir = self.download_weights()
        with open(model_dir / self.weights["classes_file"]) as infile:
            class_names = [line.strip() for line in infile.readlines()]

        self.detector = Detector(
            model_dir,
            class_names,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["max_output_size_per_class"],
            self.config["max_total_size"],
            self.config["input_size"],
            self.config["iou_threshold"],
            self.config["score_threshold"],
        )

    def predict(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts the bboxes from image frame

        Args:
            image (np.ndarray): Input image frame.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains:
            - A list of detection bboxes
            - A list of human-friendly detection class names
            - A list of detection scores
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")

        return self.detector.predict_object_bbox_from_image(image)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Estimates the 3D coordinates of an object given a 2D bounding box.
"""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Uses 2D bounding boxes information to estimate 3D location.

    Inputs:
        |bboxes_data|

    Outputs:
        |obj_3D_locs_data|

    Configs:
        focal_length (:obj:`float`): **default = 1.14**. |br|
            Approximate focal length of webcam used, in metres. Example on
            measuring focal length can be found `here <https://learnopencv.com
            /approximate-focal-length-for-webcams-and-cell-phone-cameras/>`_.
        height_factor (:obj:`float`): **default = 2.5**. |br|
            A factor used to estimate real-world distance from pixels, based on
            average human height in metres. The value varies across different
            camera set-ups, and calibration may be required. Please refer to
            the :ref:`Social Distancing use case
            <use_case_social_distancing_using_object_detection>` for more
            information.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts 2D bounding boxes into 3D locations."""
        locations = []

        for bbox in inputs["bboxes"]:
            # Subtraction is to make the camera the origin of the coordinate
            # system
            center_2d = ((bbox[0:2] + bbox[2:4]) * 0.5) - np.array([0.5, 0.5])
            bbox_height = bbox[3] - bbox[1]

            z_coord = (self.focal_length * self.height_factor) / bbox_height
            x_coord = (center_2d[0] * self.height_factor) / bbox_height
            y_coord = (center_2d[1] * self.height_factor) / bbox_height

            point = np.array([x_coord, y_coord, z_coord])
            locations.append(point)

        outputs = {"obj_3D_locs": locations}

        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws large bounding boxes with tags, over identified groups of bounding boxes.
"""

from typing import Any, Dict, List

import numpy as np

from peekingduck.pipeline.nodes.draw.utils.bbox import draw_bboxes, draw_tags
from peekingduck.pipeline.nodes.draw.utils.constants import TOMATO
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Draws large bounding boxes with tags over multiple object bounding boxes
    which have been identified as belonging to the same group.

    The :term:`large_groups` data type from :mod:`dabble.check_large_groups`,
    and the ``groups`` key of the :term:`obj_attrs` data type from
    :mod:`dabble.group_nearby_objs`, are inputs for this node which identifies
    the different groups, and the group associated with each bounding box.

    For better understanding, refer to
    the :doc:`Group Size Checking use case </use_cases/group_size_checking>`.

    Inputs:
        |img_data|

        |bboxes_data|

        |obj_attrs_data|
        :mod:`draw.group_bbox_and_tag` requires the ``groups`` attribute from
        :mod:`dabble.group_nearby_objs`.

        |large_groups_data|

    Outputs:
        |none_output_data|

    Configs:
        tag (:obj:`str`): **default = "LARGE GROUP!"**. |br|
            The string message printed when a large group is detected.

    .. versionchanged:: 1.2.0
        :mod:`draw.group_bbox_and_tag` used to take in ``obj_tags``
        (:obj:`List[str]`) as an input data type, which has been deprecated and
        now subsumed under :term:`obj_attrs`. The same attribute is accessed by
        using the ``groups`` key of :term:`obj_attrs`.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws large bounding boxes over multiple object bounding boxes
        which have been identified as belonging to large groups.

        Args:
            inputs (dict): Dictionary with keys "img", "bboxes", "obj_attrs",
                "large_groups".

        Returns:
            outputs (dict): Dictionary with keys "none".
        """
        group_bboxes = self._get_group_bbox_coords(
            inputs["large_groups"], inputs["bboxes"], inputs["obj_attrs"]["groups"]
        )
        group_tags = self._get_group_tags(inputs["large_groups"], self.tag)

        # show labels set to False to reduce clutter on display
        draw_bboxes(inputs["img"], group_bboxes, [], False, TOMATO)
        draw_tags(inputs["img"], group_bboxes, group_tags, TOMATO)

        return {}

    @staticmethod
    def _get_group_bbox_coords(
        large_groups: List[int], bboxes: List[np.ndarray], obj_groups: List[int]
    ) -> List[np.ndarray]:
        """For bboxes that belong to the same large group, get the coordinates of
        a large bbox that combines all these individual bboxes. Repeat for all large
        groups.
        """
        group_bboxes = []
        bboxes = np.array(bboxes)
        obj_groups = np.array(obj_groups)
        for group in large_groups:
            # filter relevant bboxes, select top-left and bot-right corners
            group_bbox = np.array([1.0, 1.0, 0.0, 0.0])
            selected_bboxes = bboxes[obj_groups == group]
            group_bbox[:2] = np.amin(selected_bboxes, axis=0)[:2]
            group_bbox[2:] = np.amax(selected_bboxes, axis=0)[2:]
            group_bboxes.append(group_bbox)

        return group_bboxes

    @staticmethod
    def _get_group_tags(large_groups: List[int], tag: str) -> List[str]:
        """Creates a list of tags to be used for the draw_tags function."""
        group_tags = [tag for _ in range(len(large_groups))]

        return group_tags



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mosaics area bounded by bounding boxes over detected object
"""

from typing import Any, Dict, List

import cv2
import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Mosaics areas bounded by bounding boxes on image.

    The :mod:`draw.mosaic_bbox` node helps to anonymize detected objects by
    pixelating the areas bounded by bounding boxes in an image.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |img_data|

    Configs:
        mosaic_level (:obj:`int`): **default = 7**. |br|
            Defines the resolution of a mosaic filter (width |times| height).
            The number corresponds to the number of rows and columns used to
            create a mosaic. For example, the default setting
            (``mosaic_level = 7``) creates a :math:`7 \\times 7` mosaic filter.
            Increasing the number increases the intensity of pixelization over
            an area.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.mosaic_level = self.config["mosaic_level"]

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        mosaic_img = self._mosaic_bbox(inputs["img"], inputs["bboxes"])
        outputs = {"img": mosaic_img}
        return outputs

    def _mosaic_bbox(self, image: np.ndarray, bboxes: List[np.ndarray]) -> np.ndarray:
        """Mosaics areas bounded by bounding boxes on ``image``.

        Args:
            image (np.ndarray): Image in numpy array.
            bboxes (List[np.ndarray]): numpy array of detected bboxes

        Returns:
            (np.ndarray): Image with mosaicked bounding box regions.
        """
        height, width = image.shape[:2]
        # Prevent calculating mosaic on a mosaicked area
        original_image = image.copy()

        for bbox in bboxes:
            # bbox can contain negative values sometimes, ensures the ROI
            # selection is within bounds and without wrapping
            rows = slice(int(max(0, bbox[1]) * height), int(min(1, bbox[3]) * height))
            cols = slice(int(max(0, bbox[0]) * width), int(min(1, bbox[2]) * width))

            image[rows, cols] = self._mosaic(original_image[rows, cols])

        return image

    def _mosaic(self, image: np.ndarray) -> np.ndarray:
        """Mosaics a given input image.

        Args:
            image (np.ndarray): Image in numpy array.

        Returns:
            (np.ndarray): Mosaicked image in numpy array.
        """
        height, width = image.shape[:2]
        image = cv2.resize(
            image,
            (self.mosaic_level, self.mosaic_level),
            interpolation=cv2.INTER_LANCZOS4,
        )
        image = cv2.resize(image, (width, height), interpolation=cv2.INTER_NEAREST)

        return image



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PoseNet model with model types: mobilenet50, mobilenet75, mobilenet100 and
resnet.
"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.predictor import Predictor


class PoseNetModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """PoseNet model with model types: mobilenet50, mobilenet75, mobilenet100
    and resnet.
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_valid_choice("model_type", {50, 75, 100, "resnet"})
        self.check_bounds("score_threshold", "[0, 1]")

        model_dir = self.download_weights()
        self.predictor = Predictor(
            model_dir,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["model_nodes"],
            self.config["resolution"],
            self.config["max_pose_detection"],
            self.config["score_threshold"],
        )

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """ Predict poses from input frame

        Args:
            frame (np.array): image in numpy array

        Returns:
            bboxes, keypoints, keypoint_scores, keypoint_masks, keypoint_conns
            (Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]): \
            tuple containing list of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        assert isinstance(frame, np.ndarray)

        return self.predictor.predict(frame)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EfficientDet model with model types: D0-D4
"""

import json
import logging
from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.detector import (
    Detector,
)


class EfficientDetModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """EfficientDet model with model types: D0-D4"""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_valid_choice("model_type", {0, 1, 2, 3, 4})
        self.check_bounds("score_threshold", "[0, 1]")

        model_dir = self.download_weights()
        classes_path = model_dir / self.weights["classes_file"]
        class_names = {
            val["id"] - 1: val["name"]
            for val in json.loads(classes_path.read_text()).values()
        }
        self.detect_ids = config["detect_ids"]
        self.detector = Detector(
            model_dir,
            class_names,
            self.detect_ids,
            self.config["model_type"],
            self.config["num_classes"],
            self.weights["model_file"],
            self.config["model_nodes"],
            self.config["image_size"],
            self.config["score_threshold"],
        )

    @property
    def detect_ids(self) -> List[int]:
        """The list of selected object category IDs."""
        return self._detect_ids

    @detect_ids.setter
    def detect_ids(self, ids: List[int]) -> None:
        if not isinstance(ids, list):
            raise TypeError("detect_ids has to be a list")
        if not ids:
            self.logger.info("Detecting all EfficientDet classes")
        self._detect_ids = ids

    def predict(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """predict the bbox from frame

        Args:
            image (np.ndarray): Input image frame.

        Returns:
            object_bboxes(List[Numpy ndarray]): list of bboxes detected
            object_labels(List[str]): list of index labels of the
                object detected for the corresponding bbox
            object_scores(List[float]): list of confidence scores of the
                object detected for the corresponding bbox
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")

        # returns object_bboxes, object_labels, object_scores
        return self.detector.predict_object_bbox_from_image(image)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main class for HRNet Model."""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.detector import Detector


class HRNetModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """HRNet model to detect poses from detected bboxes."""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds("score_threshold", "[0, 1]")

        model_dir = self.download_weights()
        self.detector = Detector(
            model_dir,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["model_nodes"],
            self.config["resolution"],
            self.config["score_threshold"],
        )

    def predict(
        self, frame: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts poses from input frame and bboxes.

        Args:
            frame (np.ndarray): Image in numpy array.
            bboxes (np.ndarray): Detected bboxes in image.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Tuple containing list
            of pose related info, i.e., coordinates, scores, and connections.
        """
        assert isinstance(frame, np.ndarray)
        assert isinstance(bboxes, np.ndarray)
        detected_bboxes = bboxes.copy()
        if bboxes.size != 0:
            return self.detector.predict(frame, detected_bboxes)
        keypoints = np.array([])
        keypoint_scores = np.array([])
        keypoint_conns = np.array([])

        return keypoints, keypoint_scores, keypoint_conns



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pragma: no cover

"""
Node template for creating custom nodes.
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """This is a template class of how to write a node for PeekingDuck.

    Args:
        config (:obj:`Dict[str, Any]` | :obj:`None`): Node configuration.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        # initialize/load any configs and models here
        # configs can be called by self.<config_name> e.g. self.filepath
        # self.logger.info(f"model loaded with configs: config")

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:  # type: ignore
        """This node does ___.

        Args:
            inputs (dict): Dictionary with keys "__", "__".

        Returns:
            outputs (dict): Dictionary with keys "__".
        """

        # result = do_something(inputs["in1"], inputs["in2"])
        # outputs = {"out1": result}
        # return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Custom PNG reader to fix opencv 'PNG magic' problem on Windows platform
"""
from typing import Any, Tuple
import cv2
import numpy as np


class PNGReader:
    """Custom PNG reader to fix opencv 'PNG magic' problem on Windows platform"""

    def __init__(self, input_source: str) -> None:
        self.img = cv2.imread(input_source)
        self.height, self.width, _ = self.img.shape
        self.get_map = {
            cv2.CAP_PROP_FPS: 0,
            cv2.CAP_PROP_FRAME_COUNT: 1,
            cv2.CAP_PROP_FRAME_WIDTH: self.width,
            cv2.CAP_PROP_FRAME_HEIGHT: self.height,
        }
        self.has_frames: bool = True

    def get(self, param: Any) -> int:
        """To mimic opencv's video capture object get(cv2.SOME_PROPERTY)

        Args:
            param (Any): cv2 property

        Returns:
            int: value of cv2 property if supported, otherwise -1
        """
        if param in self.get_map:
            return self.get_map[param]
        return -1

    def isOpened(self) -> bool:  # pylint: disable=invalid-name, no-self-use
        """To mimic opencv's video capture object isOpened()

        Returns:
            bool: always True
        """
        return True

    def read(self) -> Tuple[bool, np.ndarray]:
        """To mimic opencv's video capture object read()

        Returns:
            Tuple[bool, np.ndarray]: tuple of return status, image frame
        """
        if self.has_frames:
            self.has_frames = False  # only 1 frame to read and return
            return True, self.img
        return False, None

    def release(self) -> None:
        """To mimic opencv's video capture object release().
        Dummy method does nothing.
        """



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python package requirements checker."""

import collections
import importlib
import logging
import subprocess
import sys
from pathlib import Path
from typing import Any, Iterator, TextIO, Tuple, Union

import pkg_resources as pkg

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
PKD_NODE_PREFIX = "peekingduck.pipeline.nodes."
PKD_REQ_TYPE_LEN = 6  # string length of either PYTHON or SYSTEM
PKD_REQ_TYPE_PYTHON = "PYTHON"  # type specifier for Python packages
ROOT = Path(__file__).resolve().parents[1]

OptionalRequirement = collections.namedtuple("OptionalRequirement", "name type")


class RequirementChecker(importlib.abc.MetaPathFinder):
    """Checks for optional requirements from imports.

    While inheriting from MetaPathFinder is not strictly necessary, it serves
    as a reference for the required interface.
    """

    n_update = 0

    @staticmethod
    def find_spec(fullname: str, *_: Any) -> None:
        """Checks if the peekingduck.pipeline.nodes module being imported
        contains optional requirements. Attempt to install if it does.

        Args:
            fullname (:obj:`str`): Name of the module being imported.
        """
        if fullname.startswith(PKD_NODE_PREFIX):
            try:
                RequirementChecker.n_update += check_requirements(
                    fullname[len(PKD_NODE_PREFIX) :]
                )
            except subprocess.CalledProcessError:
                sys.exit(1)


def check_requirements(
    identifier: str, requirements_path: Path = ROOT / "optional_requirements.txt"
) -> int:
    """Checks if the packages specified by the ``identifier`` in the
    requirements file at ``requirements_path`` are present on the system. If
    ``install`` is ``True``, attempts to install the packages.

    Args:
        identifier (:obj:`str`): A unique identifier, typically a pipeline node
            name, used to specify which packages to check for.
        requirements_path (Path): Path to the requirements file

    Returns:
        (:obj:`int`): The number of packages updated.
    """
    with open(requirements_path) as infile:
        requirements = list(_parse_requirements(infile, identifier))

    n_update = 0
    for req in requirements:
        if req.type == PKD_REQ_TYPE_PYTHON:
            try:
                pkg.require(req.name)
            except (pkg.DistributionNotFound, pkg.VersionConflict):
                logger.info(
                    f"{req.name} not found and is required, attempting auto-update..."
                )
                try:
                    logger.info(
                        subprocess.check_output(["pip", "install", req.name]).decode()
                    )
                    n_update += 1
                except subprocess.CalledProcessError as exception:
                    logger.error(exception)
                    raise
        else:
            logger.warning(
                f"The {identifier} node requires {req.name.strip()} which needs to be "
                "manually installed. Please follow the instructions at "
                "https://peekingduck.readthedocs.io/en/stable/master.html#api-documentation "
                "and rerun. Ignore this warning if the package is already installed"
            )

    return n_update


def _parse_requirements(file: TextIO, identifier: str) -> Iterator[OptionalRequirement]:
    """Yield ``OptionalRequirement`` objects for each specification in
    ``strings``.

    ``strings`` must be a string, or a (possibly-nested) iterable thereof.

    Arg:
        file (TextIO): The file object containing optional requirements.
        identifier (str): A unique identifier, typically a pipeline node name,
            used to specify which packages to check for.

    Returns:
        (Iterator[OptionalRequirements]): Optional requirements, both Python
            and system packages, specified under the unique identifier.
    """
    lines = iter(_yield_lines(file, identifier))
    for line in lines:
        # Drop comments -- a hash without a space may be in a URL.
        if " #" in line:
            line = line[: line.find(" #")]
        req_type, req_name = _split_type_and_name(line)
        if req_type == PKD_REQ_TYPE_PYTHON:
            req = pkg.Requirement(req_name)  # type: ignore
            requirement = OptionalRequirement(f"{req.name}{req.specifier}", req_type)
        else:
            requirement = OptionalRequirement(req_name, req_type)
        yield requirement


def _yield_lines(strings: Union[TextIO, str], identifier: str) -> Iterator[str]:
    """Yield lines with ``identifier`` as the prefix.

    Args:
        strings (Union[TextIO, str]): Either a file object or a line from the
            file.
        identifier (str): A unique identifier, typically a pipeline node name,
            used to specify which packages to check for.

    Returns:
        (Iterator[str]): Lines with ``identifier`` as the prefix.
    """
    prefix = f"{identifier} "
    if isinstance(strings, str):
        for string in strings.splitlines():
            string = string.strip()
            # Return only optional requirement lines
            if string and string.startswith(prefix):
                yield string[len(prefix) :]
    else:
        for string_item in strings:
            for string in _yield_lines(string_item, identifier):
                yield string


def _split_type_and_name(string: str) -> Tuple[str, str]:
    """Split an optional requirement line into the requirement type and
    name.
    """
    return string[:PKD_REQ_TYPE_LEN], string[PKD_REQ_TYPE_LEN:]



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Checks if number of objects in a group exceeds a threshold.
"""

from collections import Counter
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Checks which groups have exceeded the group size threshold. The group
    associated with each object is accessed by the ``groups`` key of
    :term:`obj_attrs`.

    Inputs:
        |obj_attrs_data|
        :mod:`dabble.check_large_groups` requires the ``groups`` attribute.

    Outputs:
        |large_groups_data|

    Configs:
        group_size_threshold (:obj:`int`): **default = 5**. |br|
            Threshold of group size.

    .. versionchanged:: 1.2.0
        :mod:`draw.check_large_groups` used to take in ``obj_tags``
        (:obj:`List[str]`) as an input data type, which has been deprecated and
        now subsumed under :term:`obj_attrs`. The same attribute is accessed by
        using the ``groups`` key of :term:`obj_attrs`.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, List[int]]:
        """Checks which groups have exceeded the group size threshold,
        and returns a list of such groups.
        """
        group_counter: Counter = Counter(inputs["obj_attrs"]["groups"])
        large_groups = [
            group
            for group in group_counter
            if group_counter[group] > self.group_size_threshold
        ]

        return {"large_groups": large_groups}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Shows the outputs on your display.
"""

from typing import Any, Dict

import cv2

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Streams the output on your display.

    Inputs:
        |img_data|

    Outputs:
        |pipeline_end_data|

    Configs:
        window_name (:obj:`str`): **default = "PeekingDuck"** |br|
            Name of the displayed window.
        window_size (:obj:`Dict`):
            **default = { do_resizing: False, width: 1280, height: 720 }** |br|
            Resizes the displayed window to the chosen width and weight, if
            ``do_resizing`` is set to ``true``. The size of the displayed
            window can also be adjusted by clicking and dragging.
        window_loc (:obj:`Dict`): **default = { x: 0, y: 0 }** |br|
            X and Y coordinates of the top left corner of the displayed window,
            with reference from the top left corner of the screen, in pixels.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        cv2.namedWindow(self.window_name, cv2.WINDOW_NORMAL)
        cv2.moveWindow(self.window_name, self.window_loc["x"], self.window_loc["y"])

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Show the outputs on your display"""

        img = inputs["img"]
        if self.window_size["do_resizing"]:
            img = cv2.resize(
                img, (self.window_size["width"], self.window_size["height"])
            )
        cv2.imshow(self.window_name, img)

        if cv2.waitKey(1) & 0xFF == ord("q"):
            cv2.destroyWindow(self.window_name)
            return {"pipeline_end": True}

        return {"pipeline_end": False}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads inputs from multiple visual sources |br|
- image or video file on local storage |br|
- folder of images or videos |br|
- online cloud source |br|
- CCTV or webcam live feed
"""

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from peekingduck.pipeline.nodes.input.utils.preprocess import resize_image
from peekingduck.pipeline.nodes.input.utils.read import VideoNoThread, VideoThread
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class SourceType:  # pylint: disable=too-few-public-methods
    """Enumerated object to store input type"""

    DIRECTORY = 0
    FILE = 1
    URL = 2
    WEBCAM = 3


class Node(AbstractNode):  # pylint: disable=too-many-instance-attributes
    r"""Receives visual sources as inputs.

    Inputs:
        |none_input_data|

    Outputs:
        |img_data|

        |filename_data|

        |pipeline_end_data|

        |saved_video_fps_data|

    Configs:
        filename (:obj:`str`): **default = "video.mp4"**. |br|
            If source is a live stream/webcam, filename defines the name of the
            MP4 file if the media is exported. |br|
            If source is a local file or directory of files, then filename is
            the current file being processed, and the value specified here is
            overridden.
        mirror_image (:obj:`bool`): **default = False**. |br|
            Flag to set extracted image frame as mirror image of input stream.
        resize (:obj:`Dict`):
            **default = { do_resizing: False, width: 1280, height: 720 }** |br|
            Dimension of extracted image frame.
        source (:obj:`Union[int, str]`):
            **default = https://storage.googleapis.com/peekingduck/videos/wave.mp4**. |br|
            Input source can be: |br|
            - filename : local image or video file |br|
            - directory name : all media files will be processed |br|
            - http URL for online cloud source : http[s]://... |br|
            - rtsp URL for CCTV : rtsp://... |br|
            - 0 for webcam live feed |br|
            Refer to `OpenCV documentation
            <https://docs.opencv.org/4.5.5/d8/dfe/classcv_1_1VideoCapture.html>`_
            for more technical information.

        frames_log_freq (:obj:`int`): **default = 100**. [#]_ |br|
            Logs frequency of frames passed in CLI
        saved_video_fps (:obj:`int`): **default = 10**. [1]_ |br|
            This is used by :mod:`output.media_writer` to set the FPS of the
            output file and its behavior is determined by the type of input
            source. |br|
            If source is an image file, this value is ignored as it is not
            applicable. |br|
            If source is a video file, this value will be overridden by the
            actual FPS of the video. |br|
            If source is a live stream/webcam, this value is used as the FPS of
            the output file.  It is recommended to set this to the actual FPS
            obtained on the machine running PeekingDuck
            (using :mod:`dabble.fps`).
        threading (:obj:`bool`): **default = False**. [1]_ |br|
            Flag to enable threading when reading frames from camera / live
            stream. The FPS can increase up to 30%. |br|
            There is no need to enable threading if reading from a video file.
        buffering (:obj:`bool`): **default = False**. [1]_ |br|
            Boolean to indicate if threaded class should buffer image frames.
            If reading from a video file and threading is True, then buffering
            should also be True to avoid "lost frames": which happens when the
            video file is read faster than it is processed.
            One side effect of setting threading=True, buffering=True for a
            live stream/webcam is the onscreen video could appear to be playing
            in slow-mo.

    .. [#] advanced configuration

    **Technotes:**

    The following table summarizes the combinations of threading and buffering:

    +---------------------------------------+------------+--------------+
    | **Threading**                         |   False    |     True     |
    +---------------------------------------+------------+-------+------+
    | **Buffering**                         | False/True | False | True |
    +-----------+---------------------------+------------+-------+------+
    |           | Image file                |     Ok     |   Ok  |  Ok  |
    |           +---------------------------+------------+-------+------+
    |**Sources**| Video file                |     Ok     |   !   |  Ok  |
    |           +---------------------------+------------+-------+------+
    |           | Webcam, http/rtsp stream  |     Ok     |  \+   |  !!  |
    +-----------+---------------------------+------------+-------+------+

    Table Legend:

    Ok : normal behavior |br|
    \+ : potentially faster FPS |br|
    ! : lost frames if source is faster than PeekingDuck |br|
    !! : "slow-mo" video, potential out-of-memory error due to buffer overflow
    if source is faster than PeekingDuck

    Note: If threading=False, then the secondary parameter buffering is ignored
    regardless if it is set to True/False.

    Here is a video to illustrate the differences between
    `a normal video vs a "slow-mo" video
    <https://storage.googleapis.com/peekingduck/videos/wave_normal_vs_laggy.mp4>`_
    using a 30 FPS webcam: the video on the right appears to be playing in slow
    motion compared to the normal video on the left.
    This happens as both threading and buffering are set to True, and the
    threaded :mod:`input.visual` reads the webcam at almost 60 FPS.
    Since the hardware is physically limited at 30 FPS, this means every frame
    gets duplicated, resulting in each frame being processed and shown twice,
    thus "stretching out" the video.
    """

    def __init__(
        self,
        config: Dict[str, Any] = None,
        node_path: str = "",
        pkd_base_dir: Optional[Path] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._image_ext = ["gif", "jpeg", "jpg", "png"]
        self._video_ext = ["avi", "m4v", "mkv", "mov", "mp4"]
        self._allowed_extensions = self._image_ext + self._video_ext
        self._fps: float = 0  # self._fps > 0 if file playback
        self._file_name: str = ""
        self._filepaths: List[Path] = []
        self.do_resize: bool = self.resize["do_resizing"]
        self.frame_counter: int = 0
        self.total_frame_count: int = 0
        self.has_multiple_inputs: bool = False
        self.progress: int = 0
        self.videocap: Optional[Union[VideoNoThread, VideoThread]] = None
        self._determine_source_type()
        # error checking for user-defined output filename
        if not self._is_valid_file_type(Path(self.filename)):
            raise ValueError(
                f"filename {self.filename}: extension must be one of {self._allowed_extensions}"
            )
        self._open_next_input()

    def release_resources(self) -> None:
        """Override base class method to free video resource"""
        if self.videocap:
            self.videocap.shutdown()

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        outputs = self._get_next_frame()
        if self.file_end and self.has_multiple_inputs:
            self.logger.info(
                f"Completed processing file: {self._file_name}"
                f" ({self._curr_file_num} / {self._num_files})"
            )
            self.logger.debug(f"#frames={self.frame_counter}, done={self.progress}%")
            self._open_next_input()
            outputs = self._get_next_frame()
        return outputs

    def _determine_source_type(self) -> None:
        """
        Determine which one of the following types is self.source:
            - directory of files
            - file
            - url : http / rtsp
            - webcam
        If input source is a directory of files,
        then node will have specific methods to handle it.
        Otherwise opencv can deal with all non-directory sources.
        """
        if isinstance(self.source, int):
            self._source_type = SourceType.WEBCAM
        elif str(self.source).startswith(("http://", "https://", "rtsp://")):
            self._source_type = SourceType.URL
        else:
            # either directory or file
            path = Path(self.source)
            if not path.exists():
                raise FileNotFoundError(f"Path '{path}' does not exist")
            if path.is_dir():
                self._source_type = SourceType.DIRECTORY
                self._get_files(Path(self.source))
                self.has_multiple_inputs = True
                self._num_files = len(self._filepaths)
                self._curr_file_num = 0
            else:
                self._source_type = SourceType.FILE
                self._file_name = path.name

    def _get_files(self, path: Path) -> None:
        """Read all files in given directory (non-recursive)

        Args:
            path (Path): the directory path

        Raises:
            FileNotFoundError: directory does not exist error
        """
        if not path.exists():
            raise FileNotFoundError("Filepath does not exist")

        self.logger.info(f"Directory: {path}")
        self._filepaths = list(path.iterdir())
        self._filepaths.sort()

    def _get_next_frame(self) -> Dict[str, Any]:
        """Read next frame from current input file/source"""
        self.file_end = True  # assume no more frames
        outputs = {
            "img": None,
            "filename": self._file_name if self._file_name else self.filename,
            "pipeline_end": True,
            "saved_video_fps": self._fps if self._fps > 0 else self.saved_video_fps,
        }
        if self.videocap:
            success, img = self.videocap.read_frame()
            if success:
                self.file_end = False
                if self.do_resize:
                    img = resize_image(img, self.resize["width"], self.resize["height"])
                outputs["img"] = img
                outputs["pipeline_end"] = False
                self._show_progress()
            else:
                self.logger.debug("No video frames available for processing.")
        return outputs

    def _is_valid_file_type(self, filepath: Path) -> bool:
        """Check if given file has a supported file extension.

        Args:
            filepath (Path): the file to be file-type checked

        Returns:
            bool: True if supported file type else False
        """
        return filepath.suffix[1:] in self._allowed_extensions

    def _open_input(self, input_source: Any) -> None:
        """Open given input source for consumption.

        Args:
            input_source (Any): any of the following supported inputs
                                - image or video file on local storage
                                - folder of images or videos
                                - online cloud source
                                - CCTV or webcam live feed
        """
        if self.threading:
            self.videocap = VideoThread(self.source, self.mirror_image, self.buffering)
        else:
            self.videocap = VideoNoThread(input_source, self.mirror_image)
        self._fps = self.videocap.fps
        self.total_frame_count = max(0, self.videocap.frame_count)
        self.frame_counter = 0  # reset for newly opened input
        self._progress_tenth: int = 1  # each 10% progress
        # check resizing configuration
        width, height = self.videocap.resolution
        self.logger.info(f"Input size: {width} by {height}")
        if self.do_resize:
            self.logger.info(
                f"Resizing of input set to {self.resize['width']} by {self.resize['height']}"
            )

    def _open_next_file(self) -> None:
        """Load next file in a directory of files"""
        while self._filepaths:
            file_path = self._filepaths.pop(0)
            self._file_name = file_path.name
            self._curr_file_num += 1
            if self._is_valid_file_type(file_path):
                self._open_input(str(file_path))
                break  # do not proceed to next file
            self.logger.warning(
                f"Skipping '{file_path}' as it is not an accepted "
                f"file format {str(self._allowed_extensions)}"
                f" ({self._curr_file_num} / {self._num_files})"
            )

    def _open_next_input(self) -> None:
        """To open the next input source"""
        if self.has_multiple_inputs:
            self._open_next_file()
        else:
            self._open_input(self.source)

    def _show_progress(self) -> None:
        """Show progress information during pipeline iteration"""
        self.frame_counter += 1
        if self.frame_counter % self.frames_log_freq == 0 and self.videocap:
            buffer_info = (
                f", buffer: {self.videocap.queue_size}"
                if self.threading and self.buffering
                else ""
            )
            self.logger.info(f"Frames Processed: {self.frame_counter}{buffer_info}")
        if self.total_frame_count > 0:
            # more accurate to round down with int() than just round()
            self.progress = int(100 * (self.frame_counter / self.total_frame_count))
            progress_tenth = self.progress // 10
            if self.total_frame_count > 1 and progress_tenth >= self._progress_tenth:
                # progress only meaningful if input has > 1 frame
                self.logger.info(f"Approximate Progress: {self.progress}%")
                self._progress_tenth += 1  # next 10% progress



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reader functions for input nodes
"""

import logging
import platform
import queue
from pathlib import Path
from threading import Event, Thread
from typing import Any, Tuple, Union

import cv2

from peekingduck.pipeline.nodes.input.utils.png_reader import PNGReader
from peekingduck.pipeline.nodes.input.utils.preprocess import mirror


class VideoThread:
    """
    Videos will be threaded to improve FPS by reducing I/O blocking latency.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self, input_source: Union[int, str], mirror_image: bool, buffering: bool
    ) -> None:
        assert isinstance(input_source, (int, str))
        if isinstance(input_source, int):
            if platform.system().startswith("Windows"):
                # to eliminate opencv's "[WARN] terminating async callback" on Windows
                self.stream = cv2.VideoCapture(input_source, cv2.CAP_DSHOW)
            else:
                self.stream = cv2.VideoCapture(input_source)
        elif Path(input_source.lower()).suffix == ".png":
            self.stream = PNGReader(input_source)
        else:
            self.stream = cv2.VideoCapture(input_source)
        if not self.stream.isOpened():
            raise ValueError(f"Camera or video input not detected: {input_source}")
        self.logger = logging.getLogger(type(self).__name__)
        self.mirror = mirror_image
        # events to coordinate threading
        self.is_done = Event()
        self.is_thread_start = Event()
        # frame storage and buffering
        self.frame_counter = 0
        self.frame = None
        self.prev_frame = None
        self.buffer = buffering
        self.queue: queue.Queue = queue.Queue()
        # start threading
        self.thread = Thread(target=self._reading_thread, args=(), daemon=True)
        self.thread.start()
        self.is_thread_start.wait()

    def __del__(self) -> None:
        """
        Release acquired resources here.
        """
        self.logger.debug("VideoThread.__del__")
        self.stream.release()

    def shutdown(self) -> None:
        """
        Shuts down this class.
        Cannot be merged into __del__ as threading code needs to run here.
        """
        self.logger.debug("VideoThread.shutdown")
        self.is_done.set()
        self.thread.join()

    def _reading_thread(self) -> None:
        """
        A thread that continuously polls the camera for frames.
        """
        while not self.is_done.is_set():
            if self.stream.isOpened():
                ret, frame = self.stream.read()
                if not ret:
                    self.logger.debug(
                        f"_reading_thread: ret={ret}, "
                        f"#frames read={self.frame_counter}"
                    )
                    self.is_done.set()
                else:
                    if self.mirror:
                        frame = mirror(frame)
                    self.frame = frame
                    self.is_thread_start.set()  # thread really started
                    self.frame_counter += 1
                    if self.buffer:
                        self.queue.put(self.frame)

    def read_frame(self) -> Tuple[bool, Any]:
        """
        Reads the frame.
        """
        # pylint: disable=no-else-return
        if self.buffer:
            if self.queue.empty():
                if self.is_done.is_set():
                    # end of input
                    return False, None
                else:
                    # input slow, so duplicate frame
                    return True, self.prev_frame
            else:
                self.prev_frame = self.queue.get()
                return True, self.prev_frame
        else:
            if self.is_done.is_set():
                return False, None
            else:
                return True, self.frame

    @property
    def fps(self) -> float:
        """Get FPS of videofile

        Returns:
            int: number indicating FPS
        """
        fps = self.stream.get(cv2.CAP_PROP_FPS)
        return fps

    @property
    def frame_count(self) -> int:
        """Get total number of frames of file

        Returns:
            int: number indicating frame count
        """
        num_frames = self.stream.get(cv2.CAP_PROP_FRAME_COUNT)
        return int(num_frames)

    @property
    def queue_size(self) -> int:
        """Get buffer queue size

        Returns:
            int: number of frames in buffer
        """
        return self.queue.qsize()

    @property
    def resolution(self) -> Tuple[int, int]:
        """Get resolution of the camera device used.

        Returns:
            width(int): width of input resolution
            height(int): heigh of input resolution
        """
        width = self.stream.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
        return int(width), int(height)


class VideoNoThread:
    """
    No threading to deal with recorded videos and images.
    """

    def __init__(self, input_source: Union[int, str], mirror_image: bool) -> None:
        assert isinstance(input_source, (int, str))
        if isinstance(input_source, int):
            if platform.system().startswith("Windows"):
                # to eliminate opencv's "[WARN] terminating async callback" on Windows
                self.stream = cv2.VideoCapture(input_source, cv2.CAP_DSHOW)
            else:
                self.stream = cv2.VideoCapture(input_source)
        elif Path(input_source.lower()).suffix == ".png":
            self.stream = PNGReader(input_source)
        else:
            self.stream = cv2.VideoCapture(input_source)
        if not self.stream.isOpened():
            raise ValueError(f"Video or image path incorrect: {input_source}")
        self._frame_counter = 0
        self.logger = logging.getLogger(type(self).__name__)
        self.mirror = mirror_image

    def __del__(self) -> None:
        # dotw: self.logger.debug below crashes on Nvidia Jetson Xavier Ubuntu 18.04 python 3.6
        #       but does not crash on Intel MacBook Pro Ubuntu 20.04 python 3.7
        # self.logger.debug("VideoNoThread.__del__")
        self.stream.release()

    def read_frame(self) -> Tuple[bool, Any]:
        """
        Reads the frame.
        """
        ret, frame = self.stream.read()
        if not ret:
            self.logger.debug(
                f"read_frame: ret={ret}, #frames read={self._frame_counter}"
            )
        else:
            self._frame_counter += 1
        return ret, frame

    # pylint: disable=R0201
    def shutdown(self) -> None:
        """
        Shuts down this class.
        Cannot be merged into __del__ as threading code needs to run here.
        Dummy method left here for consistency with VideoThread class.
        """
        self.logger.debug("VideoNoThread.shutdown")

    @property
    def fps(self) -> float:
        """Get FPS of videofile

        Returns:
            int: number indicating FPS
        """
        fps = self.stream.get(cv2.CAP_PROP_FPS)
        return fps

    @property
    def frame_count(self) -> int:
        """Get total number of frames of file

        Returns:
            int: number indicating frame count
        """
        num_frames = self.stream.get(cv2.CAP_PROP_FRAME_COUNT)
        return int(num_frames)

    @property
    def queue_size(self) -> int:
        """Get buffer queue size

        Returns:
            int: number of frames in buffer
        """
        return 0

    @property
    def resolution(self) -> Tuple[int, int]:
        """Get resolution of the file.

        Returns:
            width(int): width of resolution
            height(int): heigh of resolution
        """
        width = self.stream.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
        return int(width), int(height)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Network blocks for constructing the Darknet-53 backbone of the JDE model.

Modifications include:
- Removed custom Upsample module
- Removed EmptyLayer module (the equivalent nn.Identity is available)
- Removed training related code in YOLOLayer.forward()
- Removed loss related member variables
- Removed img_size in constructor since it's ignored and self.img_size is
    initialized to 0 by default
- Removed `layer` member variable in YOLOLayer since it's not used
- Updating self.img_size after creating grid to avoid recreation
    - Refactored initial value of self.img_size to None to indicate it's
        meant to be overwritten
"""

import math
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from peekingduck.pipeline.nodes.model.jdev1.jde_files.utils import decode_delta_map


class YOLOLayer(nn.Module):  # pylint: disable=too-many-instance-attributes
    """YOLO detection layer.

    Args:
        anchors (List[Tuple[float, float]]): List of anchor box width and
            heights.
        num_classes (int): Number of classes. Uses 1 for JDE.
        num_identities (int): Number of identities, e.g., number of unique
            pedestrians. Uses 14455 for JDE according to the original code.
        embedding_dim (int): Size of embedding. Uses 512 for JDE according to
            the original code.
        device (torch.device): The device which a `torch.Tensor` is on or
            will be allocated.
    """

    def __init__(  # pylint:disable=too-many-arguments
        self,
        anchors: List[Tuple[float, float]],
        num_classes: int,
        num_identities: int,
        embedding_dim: int,
        device: torch.device,
    ) -> None:
        super().__init__()
        self.anchors = torch.FloatTensor(anchors)
        self.num_anchors = len(anchors)  # number of anchors (4)
        self.num_classes = num_classes  # number of classes (80)
        self.num_identities = num_identities
        self.emb_dim = embedding_dim
        self.device = device
        self.img_size: Optional[Tuple[int, int]] = None
        self.shift = [1, 3, 5]

        self.anchor_vec: torch.Tensor
        self.anchor_wh: torch.Tensor
        self.grid_xy: torch.Tensor
        self.stride: float

        self.emb_scale = (
            math.sqrt(2) * math.log(self.num_identities - 1)
            if self.num_identities > 1
            else 1
        )

    def forward(self, inputs: torch.Tensor, img_size: Tuple[int, int]) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            inputs (torch.Tensor): Feature maps at various scales.
            img_size (Tuple[int, int]): Image size as specified by backbone
                configuration.

        Returns:
            (torch.Tensor): A decoded tensor containing the prediction.
        """
        # From arxiv article:
        # Prediction map has dimension B * (6A + D) * H * W where A is number
        # of anchor templates, D is embedding dimension. B, H, and W are
        # batch size, height, and width (of the feature maps) respectively.
        pred_anchor, pred_embedding = inputs[:, :24, ...], inputs[:, 24:, ...]
        batch_size, grid_height, grid_width = (
            pred_anchor.shape[0],
            pred_anchor.shape[-2],
            pred_anchor.shape[-1],
        )

        if self.img_size != img_size:
            self._create_grids(img_size, grid_height, grid_width)
            # Only have to create the grid once for each image size
            self.img_size = img_size
            if pred_anchor.is_cuda:
                self.grid_xy = self.grid_xy.cuda()
                self.anchor_wh = self.anchor_wh.cuda()

        # prediction
        pred_anchor = (
            pred_anchor.view(
                batch_size,
                self.num_anchors,
                self.num_classes + 5,
                grid_height,
                grid_width,
            )
            .permute(0, 1, 3, 4, 2)
            .contiguous()
        )

        pred_embedding = pred_embedding.permute(0, 2, 3, 1).contiguous()
        pred_box = pred_anchor[..., :4]
        pred_conf = pred_anchor[..., 4:6].permute(0, 4, 1, 2, 3)

        pred_conf = torch.softmax(pred_conf, dim=1)[:, 1, ...].unsqueeze(-1)
        pred_embedding = F.normalize(
            pred_embedding.unsqueeze(1)
            .repeat(1, self.num_anchors, 1, 1, 1)
            .contiguous(),
            dim=-1,
        )
        pred_cls = torch.zeros(
            batch_size, self.num_anchors, grid_height, grid_width, 1
        ).to(self.device)
        pred_anchor = torch.cat([pred_box, pred_conf, pred_cls, pred_embedding], dim=-1)
        pred_anchor[..., :4] = decode_delta_map(
            pred_anchor[..., :4], self.anchor_vec.to(pred_anchor), self.device
        )
        pred_anchor[..., :4] *= self.stride

        return pred_anchor.view(batch_size, -1, pred_anchor.shape[-1])

    def _create_grids(
        self, img_size: Tuple[int, int], grid_height: int, grid_width: int
    ) -> None:
        """Builds the grid for anchor box offsets.

        Args:
            img_size (Tuple[int, int]): Model input size.
            grid_height (int): Height of grid.
            grid_width (int): Width of grid.
        """
        self.stride = img_size[0] / grid_width
        assert (
            self.stride == img_size[1] / grid_height
        ), f"Inconsistent stride size: {self.stride} v.s. {img_size[1]} / {grid_height}"

        # build xy offsets
        grid_x = (
            torch.arange(grid_width)
            .repeat((grid_height, 1))
            .view((1, 1, grid_height, grid_width))
            .float()
        )
        grid_y = (
            torch.arange(grid_height)
            .repeat((grid_width, 1))
            .transpose(0, 1)
            .view((1, 1, grid_height, grid_width))
            .float()
        )
        self.grid_xy = torch.stack((grid_x, grid_y), 4)

        # build wh gains
        self.anchor_vec = self.anchors / self.stride
        self.anchor_wh = self.anchor_vec.view(1, self.num_anchors, 1, 1, 2)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module to display deprecation warnings."""

import warnings


def deprecate(message: str, level: int) -> None:
    """Issues a deprecation warning using the FutureWarning category.

    Args:
        message (str): The deprecation warning message.
        level (int): The `stacklevel`, can be used to refer to `deprecate()`'s
            caller instead of the source of `deprecate()` itself.

    """
    warnings.warn(message, category=FutureWarning, stacklevel=level)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for drawing pose keypoints and connections
"""

from typing import Any, Iterable, Tuple, Union

import cv2
import numpy as np

from peekingduck.pipeline.nodes.draw.utils.constants import (
    CHAMPAGNE,
    POINT_RADIUS,
    THICK,
    TOMATO,
)
from peekingduck.pipeline.nodes.draw.utils.general import (
    get_image_size,
    project_points_onto_original_image,
)


def draw_human_poses(
    image: np.ndarray, keypoints: np.ndarray, keypoint_conns: np.ndarray
) -> None:
    # pylint: disable=too-many-arguments
    """Draw poses onto an image frame.

    Args:
        image (np.array): image of current frame
        keypoints (List[Any]): list of keypoint coordinates
        keypoints_conns (List[Any]): list of keypoint connections
    """
    image_size = get_image_size(image)
    num_persons = keypoints.shape[0]
    if num_persons > 0:
        for i in range(num_persons):
            _draw_connections(image, keypoint_conns[i], image_size, CHAMPAGNE)
            _draw_keypoints(image, keypoints[i], image_size, TOMATO, POINT_RADIUS)


def _draw_connections(
    frame: np.ndarray,
    connections: Union[None, Iterable[Any]],
    image_size: Tuple[int, int],
    connection_color: Tuple[int, int, int],
) -> None:
    """Draw connections between detected keypoints"""
    if connections is not None:
        for connection in connections:
            pt1, pt2 = project_points_onto_original_image(connection, image_size)
            cv2.line(frame, (pt1[0], pt1[1]), (pt2[0], pt2[1]), connection_color, THICK)


def _draw_keypoints(
    frame: np.ndarray,
    keypoints: np.ndarray,
    image_size: Tuple[int, int],
    keypoint_dot_color: Tuple[int, int, int],
    keypoint_dot_radius: int,
) -> None:
    # pylint: disable=too-many-arguments
    """Draw detected keypoints"""
    img_keypoints = project_points_onto_original_image(keypoints, image_size)

    for _, keypoint in enumerate(img_keypoints):
        _draw_one_keypoint_dot(frame, keypoint, keypoint_dot_color, keypoint_dot_radius)


def _draw_one_keypoint_dot(
    frame: np.ndarray,
    keypoint: np.ndarray,
    keypoint_dot_color: Tuple[int, int, int],
    keypoint_dot_radius: int,
) -> None:
    """Draw single keypoint"""
    cv2.circle(
        frame, (keypoint[0], keypoint[1]), keypoint_dot_radius, keypoint_dot_color, -1
    )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pipeline class that stores nodes and manages the data information used during
inference.
"""

import textwrap
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Pipeline:  # pylint: disable=too-few-public-methods
    """Pipeline class that stores nodes and manages flow of data used during
    inference.

    Args:
        nodes (:obj:`List[AbstractNode]`): List of initialized nodes for the
            pipeline to run through.
    """

    def __init__(self, nodes: List[AbstractNode]) -> None:
        self.nodes = nodes
        self._check_pipe(nodes)
        self.data = {}  # type: ignore
        self.terminate = False

    def get_pipeline_results(self) -> Dict[str, Any]:
        """Gets all results data from nodes in pipeline.

        Returns:
            (:obj:`Dict[str, Any]`): Dictionary of all results from nodes in
            the pipeline.
        """
        return self.data

    @staticmethod
    def _check_pipe(nodes: List[AbstractNode]) -> None:
        # 1. Check the initial node is a source node
        # 2. Check every subsequent node utilizes something that will exist
        # if reached end, it is all valid
        data_pool = []

        if nodes[0].inputs[0] == "none":
            data_pool.extend(nodes[0].outputs)

        for node in nodes[1:]:
            if all(item in data_pool for item in node.inputs) or "all" in node.inputs:
                data_pool.extend(node.outputs)
            else:
                msg = textwrap.dedent(
                    f"""\
                    Nodes in this pipeline do not form a proper channel:
                    {node.name} requires these inputs: {node.inputs}
                    Data pool only has these outputs from previous nodes: {data_pool}
                    Note that nodes run sequentially, in the order specified in the config file.
                    """
                )
                raise ValueError(msg)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""YOLO-based face detection model with model types: v4 and v4tiny."""

import logging
from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.yolov4_face.yolo_face_files.detector import (
    Detector,
)


class YOLOFaceModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """YOLO face model with model types: v4 and v4tiny."""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds(["iou_threshold", "score_threshold"], "[0, 1]")

        model_dir = self.download_weights()
        with open(model_dir / self.weights["classes_file"]) as infile:
            class_names = [line.strip() for line in infile.readlines()]

        self.detect_ids = config["detect_ids"]
        self.detector = Detector(
            model_dir,
            class_names,
            self.detect_ids,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["max_output_size_per_class"],
            self.config["max_total_size"],
            self.config["input_size"],
            self.config["iou_threshold"],
            self.config["score_threshold"],
        )

    @property
    def detect_ids(self) -> List[int]:
        """The list of selected object category IDs."""
        return self._detect_ids

    @detect_ids.setter
    def detect_ids(self, ids: List[int]) -> None:
        if not isinstance(ids, list):
            raise TypeError("detect_ids has to be a list")
        self._detect_ids = ids

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, labels and scores

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            labels (np.ndarray): numpy array of class labels
            scores (np.ndarray): numpy array of confidence scores
        """
        assert isinstance(frame, np.ndarray)

        return self.detector.predict_object_bbox_from_image(frame)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2017 TU Berlin, Communication Systems Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Data class to store information of a single tracked detection."""

import numpy as np


class Track:  # pylint: disable=too-few-public-methods
    """Stores information for each tracked detection.

    Args:
        track_id (int): Tracking ID of the detection.
        bbox (np.ndarray): Bounding box coordinates with (t, l, w, h) format
            where (t, l) is the top-left corner, w is the width, and h is the
            height.

    Attributes:
        bbox (np.ndarray): Bounding box coordinates with (t, l, w, h) format
            where (t, l) is the top-left corner, w is the width, and h is the
            height.
        iou_score (float): The Intersection-over-Union value between the
            current `bbox` and the immediate previous `bbox`.
        lost (int): The number of consecutive frames where this detection is
            not found in the video frame.
        track_id (int): Tracking ID of the detection.
    """

    def __init__(self, track_id: int, bbox: np.ndarray) -> None:
        self.track_id = track_id
        self.lost = 0
        self.update(bbox)

    def update(self, bbox: np.ndarray, iou_score: float = 0.0) -> None:
        """Updates the tracking result with information from the latest frame.

        Args:
            bbox (np.ndarray): Bounding box with format (t, l, w, h) where
                (t, l) is the top-left corner, w is the width, and h is the
                height.
            iou_score (float): Intersection-over-Union between the current
                detection bounding box and its last detected bounding box.
        """
        self.bbox = bbox
        self.iou_score = iou_score
        self.lost = 0



# Copyright 2018 Ross Wightman
#
# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core functions to use posenet to detect poses
"""


from typing import Callable, List

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.decode_multi import (
    decode_multiple_poses,
)


def get_keypoints_relative_coords(
    keypoint_coords: np.ndarray, output_scale: np.ndarray, image_size: List[int]
) -> np.ndarray:
    """Get keypoints coordinates relative to image size
    Args:
        keypoints_coords (np.array): Nx17x2 keypoints coordinates of N persons
        output_scale (np.array): output scale in Hx2 format
        image_size (List[int]): image size in HxW format
    Returns:
        keypoints_coords (np.array): Nx17x2 keypoints coordinates of N persons
                relative to image size
    """
    assert len(keypoint_coords.shape) == 3, "keypoint_coords should be 3D"
    assert keypoint_coords.shape[
        2
    ], "keypoint_coords should be a 2D matrix of 2D offsets"
    rel_keypoint_coords = keypoint_coords * output_scale
    rel_keypoint_coords = rel_keypoint_coords / image_size
    return rel_keypoint_coords


def _sigmoid(array: np.ndarray) -> np.ndarray:
    return 1 / (1 + np.exp(-array))


def detect_keypoints(
    tf_model: Callable,
    image: tf.Tensor,
    output_stride: int,
    dst_scores: np.ndarray,
    dst_keypoints: np.ndarray,
    model_type: str,
    score_threshold: float,
) -> int:
    # pylint: disable=too-many-arguments
    """Evaluate image by model function to get detected keypoints
    Args:
        tf_model: tensorflow model
        image (tf.Tensor): image for inference
        output_stride (int): output stride to convert output indices to image coordinates
        dst_scores (np.array): Nx17 buffer to store keypoint scores where N is
            the max persons to be detected
        dst_keypoints (np.array): Nx17x2 buffer to store keypoints coordinate
            where N is the max persons to be detected
        model_type (str): specified model type (refer to modelconfig.yml)
        score_threshold (float): threshold for prediction
    Returns:
        pose_count (int): number of poses detected
    """
    model_output = tf_model(image)

    # For resnet's implementation, we need to apply a sigmoid function on
    # the heatmap, which is the first tensor in the output
    if model_type == "resnet":
        model_output[0] = _sigmoid(model_output[0])

    pose_count = decode_multiple_poses(
        model_output,
        dst_scores,
        dst_keypoints,
        output_stride=output_stride,
        min_pose_score=score_threshold,
    )

    return pose_count



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tracking-by-detection using OpenCV's MOSSE Tracker."""

from typing import Any, Dict, List, NamedTuple

import cv2
import numpy as np

from peekingduck.pipeline.nodes.dabble.trackingv1.tracking_files.utils import (
    iou_candidates,
    xyxyn2tlwh,
)


class OpenCVTracker:  # pylint: disable=too-few-public-methods
    """OpenCV's MOSSE tracker.

    Attributes:
        is_initialized (bool): Flag to determine this is the first run of the
            tracker.
        iou_threshold (float): Minimum IoU value to be used with the matching
            logic.
        next_track_id (int): ID for the next (unmatched) object detection to be
            tracked.
        tracks (Dict[int, Track]): Maps track IDs to their respective
            tracker and bbox coordinates.
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        self.is_initialized = False
        self.iou_threshold = config["iou_threshold"]
        self.next_track_id = 0
        self.tracks: Dict[int, Track] = {}

    def track_detections(self, inputs: Dict[str, Any]) -> List[int]:
        """Initializes and updates the tracker on each frame.

        Args:
            inputs (Dict[str, Any]): Dictionary with keys "img" and "boxes".

        Returns:
            (List[int]): Tracking IDs of the detection bounding boxes.
        """
        frame = inputs["img"]
        frame_size = frame.shape[:2]
        tlwhs = xyxyn2tlwh(inputs["bboxes"], *frame_size)

        if self.is_initialized:
            obj_track_ids = self._match_and_track(frame, tlwhs)
        else:
            for tlwh in tlwhs:
                self._initialize_tracker(frame, tlwh)
            obj_track_ids = list(self.tracks.keys())
            self.is_initialized = True
        self._update_tracker_bboxes(frame)

        return obj_track_ids

    def _initialize_tracker(self, frame: np.ndarray, bbox: np.ndarray) -> None:
        """Starts a tracker for each bbox.

        Args:
            frame (np.ndarray): Image frame parsed from video.
            bbox (np.ndarray): Single detected bounding box.
        """
        tracker = cv2.legacy.TrackerMOSSE_create()
        tracker.init(frame, tuple(bbox))
        self.tracks[self.next_track_id] = Track(tracker, bbox)
        self.next_track_id += 1

    def _match_and_track(self, frame: np.ndarray, bboxes: np.ndarray) -> List[int]:
        """Matches detections to tracked bboxes, creates a new track if no
        match is found.

        Args:
            frame (np.ndarray): Input video frame.
            bboxes (np.ndarray): Detection bounding boxes with (t, l, w, h)
                format where (t, l) is the coordinate of the top-left corner,
                w is the width, and h is the height of the bounding box
                respectively.

        Returns:
            (List[int]): A list of track IDs for the detections in the current
                frame.
        """
        prev_tracked_bboxes = [track.bbox for _, track in self.tracks.items()]
        matching_dict = {}

        for bbox in bboxes:
            ious = iou_candidates(bbox, np.array(prev_tracked_bboxes))
            matching_dict[tuple(bbox)] = (
                ious.argmax() if max(ious) >= self.iou_threshold else None
            )

        track_ids = []
        for bbox, matched_id in matching_dict.items():
            if matched_id is None:
                self._initialize_tracker(frame, np.array(bbox))
                track_ids.append(list(self.tracks)[-1])
            else:
                track_ids.append(list(self.tracks)[matched_id])

        return track_ids

    def _update_tracker_bboxes(self, frame: np.ndarray) -> None:
        """Updates location of previously tracked detections in subsequent
        frames. Removes the track if its tracker fails to update.
        """
        failures = []
        for track_id, track in self.tracks.items():
            success, bbox = track.tracker.update(frame)
            if success:
                self.tracks[track_id] = Track(track.tracker, np.array(bbox))
            else:
                failures.append(track_id)
        for track_id in failures:
            del self.tracks[track_id]


class Track(NamedTuple):
    """Stores the individual OpenCV MOSSE tracker and its bounding box
    coordinates.

    Attributes:
        tracker (cv2.legacy_TrackerMOSSE): OpenCV MOSSE tracker.
        bbox (np.ndarray): Bounding box coordinates in (t, l, w, h) format where
            (t, l) is the top-left coordinate, w is the width, and h is the
            height of the bounding box.
    """

    tracker: cv2.legacy_TrackerMOSSE
    bbox: np.ndarray



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Abstract Node class for all nodes.
"""


# pylint: disable=unused-import
from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.utils.deprecation import deprecate

deprecate(
    "importing AbstractNode from peekingduck.pipeline.nodes.node is deprecated "
    "and will be removed in a future version. Please import from "
    "peekingduck.pipeline.nodes.abstract_node instead.",
    3,
)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Universal logging configuration
"""

import logging
import os
import sys
import traceback
from types import TracebackType
from typing import Dict, Optional, Type

from colorama import Fore, Style, init


class LoggerSetup:  # pylint: disable=too-few-public-methods
    """Set up the universal logging configuration"""

    def __init__(self, log_level: str = "info") -> None:
        if os.name == "nt":
            init()

        formatter = ColoredFormatter(
            "{asctime} {name} {level_color} {levelname}"
            "{reset}: {msg_color} {message} {reset}",
            style="{",
            datefmt="%Y-%m-%d %H:%M:%S",
            colors={
                "DEBUG": Fore.RESET + Style.BRIGHT,
                "INFO": Fore.RESET + Style.BRIGHT,
                "WARNING": Fore.YELLOW + Style.BRIGHT,
                "ERROR": Fore.RED + Style.BRIGHT,
                "CRITICAL": Fore.RED + Style.BRIGHT,
            },
        )

        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)

        self.logger = logging.getLogger()
        self.logger.handlers[:] = []
        self.logger.addHandler(handler)
        sys.excepthook = self.handle_exception
        LoggerSetup.set_log_level(log_level)

    def handle_exception(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> None:
        """Use Python's logging module when showing errors"""

        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        error_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
        traceback_msg = " ".join([str(elem) for elem in error_list[:-1]])
        error_msg = str(error_list[-1])

        # Make the error type more obvious in terminal by separating these
        self.logger.error(traceback_msg)
        self.logger.error(error_msg)

    @staticmethod
    def set_log_level(
        log_level: str = "info", logger_name: Optional[str] = None
    ) -> None:
        """Changes the logging level to the specified one.
        If logger_name is None, then this will change the global log level.

        Args:
            log_level (str, optional): Log level to set. Defaults to "info".
            logger_name (Optional[str], optional): Logger name. Defaults to None.
        """
        log_level_settings = set(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
        log_level = log_level.upper()
        if log_level not in log_level_settings:
            log_level = "INFO"
        logger = logging.getLogger(logger_name)
        if log_level != "INFO":
            if logger_name:
                logger.info(f"Changing log_level to {log_level}")
            else:
                logger.info(f"Changing global log_level to {log_level}")
        logger.setLevel(log_level)


class ColoredFormatter(logging.Formatter):
    """This class formats the color of logging messages"""

    def __init__(
        self, *args: str, colors: Optional[Dict[str, str]] = None, **kwargs: str
    ) -> None:
        """Initialize the formatter with specified format strings"""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record: logging.LogRecord) -> str:
        """Format the specified record as text."""

        record.level_color = self.colors.get(record.levelname, "")
        record.reset = Style.RESET_ALL
        record.msg_color = self.colors.get(record.levelname, "")

        return super().format(record)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
General utils for drawing functions
"""

from typing import Tuple

import numpy as np


def get_image_size(frame: np.ndarray) -> Tuple[int, int]:
    """Obtain image size of input frame

    Args:
        frame (np.array): image of current frame

    Returns:
        image_size (Tuple[int, int]): Width and height of image
    """
    image_size = (frame.shape[1], frame.shape[0])  # width, height
    return image_size


def project_points_onto_original_image(
    points: np.ndarray, image_size: Tuple[int, int]
) -> np.ndarray:
    """Project points from relative value (0, 1) to absolute values in original
    image. Note that coordinate (0, 0) starts from image top-left.

    Args:
        points (np.array): points on an image
        image_size (Tuple[int, int]): Width and height of image

    Returns:
        project_points (np.ndarray): projected points on the original image
                                     in integer coordinates
    """
    if len(points) == 0:
        return []

    points = points.reshape((-1, 2))

    projected_points = np.array(points, dtype=np.float32)

    width, height = image_size[0], image_size[1]
    projected_points[:, 0] *= width
    projected_points[:, 1] *= height

    return np.round(projected_points).astype(int)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🔲 High performance anchor-free YOLO object detection model."""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.model.yoloxv1 import yolox_model
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Initializes and uses YOLOX to infer from an image frame.

    The YOLOX node is capable detecting objects from 80 categories. The table
    of object categories can be found
    :ref:`here <general-object-detection-ids>`. The ``"yolox-tiny"`` model is
    used by default and can be changed to one of ``("yolox-tiny", "yolox-s",
    "yolox-m", "yolox-l")``.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

    Configs:
        model_type (:obj:`str`): **{"yolox-tiny", "yolox-s", "yolox-m",
            "yolox-l"}, default="yolox-tiny"**. |br|
            Defines the type of YOLOX model to be used.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        input_size (:obj:`int`): **default=416**. |br|
            Input image resolution of the YOLOX model.
        detect_ids (:obj:`List[int]`): **default=[0]**. |br|
            List of object category IDs to be detected. To detect all classes,
            refer to the :ref:`tech note <general-object-detection-ids>`.
        iou_threshold (:obj:`float`): **[0, 1], default = 0.45**. |br|
            Overlapping bounding boxes with Intersection over Union (IoU) above
            the threshold will be discarded.
        score_threshold (:obj:`float`): **[0, 1], default = 0.25**. |br|
            Bounding boxes with confidence score (product of objectness score
            and classification score) below the threshold will be discarded.
        agnostic_nms (:obj:`bool`): **default = True**. |br|
            Flag to determine if class-agnostic NMS (``torchvision.ops.nms``)
            or class-aware NMS (``torchvision.ops.batched_nms``) should be
            used.
        half (:obj:`bool`): **default = False**. |br|
            Flag to determine if half-precision floating-point should be used
            for inference.
        fuse (:obj:`bool`): **default = False**. |br|
            Flag to determine if the convolution and batch normalization layers
            should be fused for inference.

    References:
        YOLOX: Exceeding YOLO Series in 2021:
        https://arxiv.org/abs/2107.08430

        Inference code and model weights:
        https://github.com/Megvii-BaseDetection/YOLOX
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = yolox_model.YOLOXModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads `img` from `inputs` and return the bboxes of the detect
        objects.

        The classes of objects to be detected can be specified through the
        `detect_ids` configuration option.

        Args:
            inputs (Dict): Inputs dictionary with the key `img`.

        Returns:
            (Dict): Outputs dictionary with the keys `bboxes`, `bbox_labels`,
                and `bbox_scores`.
        """
        bboxes, labels, scores = self.model.predict(inputs["img"])
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}

        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Object detection class using YOLOv4 model to detect human faces."""

import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import cv2
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants


class Detector:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Object detection class using yolo model to find human faces."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: List[str],
        detect_ids: List[int],
        model_type: str,
        model_file: Dict[str, str],
        max_output_size_per_class: int,
        max_total_size: int,
        input_size: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.class_names = class_names
        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]

        self.max_output_size_per_class = max_output_size_per_class
        self.max_total_size = max_total_size
        self.input_size = (input_size, input_size)
        self.iou_threshold = iou_threshold
        self.score_threshold = score_threshold

        self.detect_ids = detect_ids
        self.yolo = self._create_yolo_model()

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, labels and scores.

        Args:
            image (np.ndarray): Input image.

        Returns:
            bboxes (np.ndarray): Detected bboxes
            labels (np.ndarray): Class labels for each detected bbox.
            scores (np.ndarray): Confidence scores for each detected bbox.
        """
        image = self._preprocess(image)

        pred = self.yolo(tf.constant(image))
        pred = next(iter(pred.values()))

        bboxes, scores, classes = self._postprocess(pred[:, :, :4], pred[:, :, 4:])
        labels = np.array([self.class_names[int(i)] for i in classes])

        return bboxes, labels, scores

    def _create_yolo_model(self) -> Callable:
        self.logger.info(
            "Yolo model loaded with following configs:\n\t"
            f"Model type: {self.model_type},\n\t"
            f"Input resolution: {self.input_size},\n\t"
            f"IDs being detected: {self.detect_ids},\n\t"
            f"Max detections per class: {self.max_output_size_per_class},\n\t"
            f"Max total detections: {self.max_total_size},\n\t"
            f"IOU threshold: {self.iou_threshold},\n\t"
            f"Score threshold: {self.score_threshold}"
        )

        return self._load_yolo_weights()

    def _load_yolo_weights(self) -> Callable:
        self.model = tf.saved_model.load(
            str(self.model_path), tags=[tag_constants.SERVING]
        )
        return self.model.signatures["serving_default"]

    def _postprocess(
        self,
        pred_boxes: tf.Tensor,
        pred_scores: tf.Tensor,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        bboxes, scores, classes, valid_dets = tf.image.combined_non_max_suppression(
            tf.reshape(pred_boxes, (tf.shape(pred_boxes)[0], -1, 1, 4)),
            tf.reshape(
                pred_scores, (tf.shape(pred_scores)[0], -1, tf.shape(pred_scores)[-1])
            ),
            self.max_output_size_per_class,
            self.max_total_size,
            self.iou_threshold,
            self.score_threshold,
        )
        num_valid = valid_dets[0]

        classes = classes.numpy()[0]
        classes = classes[:num_valid]
        # only identify objects we are interested in
        mask = np.isin(classes, self.detect_ids)

        scores = scores.numpy()[0]
        scores = scores[:num_valid]
        scores = scores[mask]

        bboxes = bboxes.numpy()[0]
        bboxes = bboxes[:num_valid]
        bboxes = bboxes[mask]

        # swapping x and y axes
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        return bboxes, scores, classes

    def _preprocess(self, image: np.ndarray) -> np.ndarray:
        image = cv2.resize(image, self.input_size)
        image = np.asarray([image]).astype(np.float32) / 255.0

        return image



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🔲 Scalable and efficient object detection."""

from typing import Any, Dict

import cv2
import numpy as np

from peekingduck.pipeline.nodes.model.efficientdet_d04 import efficientdet_model
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Initializes an EfficientDet model to detect bounding boxes from an image.

    The EfficientDet node is capable of detecting objects from 80 categories.
    The table of categories can be found
    :ref:`here <general-object-detection-ids>`.

    EfficientDet node has five levels of compound coefficient (0 - 4). A higher
    compound coefficient will scale up all dimensions of the backbone network
    width, depth, input resolution, feature network, and box/class prediction
    at the same time, which results in better performance but slower inference
    time. The default compound coefficient is 0 and can be changed to other
    values.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

    Configs:
        model_type (:obj:`int`): **{0, 1, 2, 3, 4}, default = 0**. |br|
            Defines the compound coefficient for EfficientDet.
        score_threshold (:obj:`float`): **[0, 1], default = 0.3**.
            Bounding boxes with confidence score below the threshold will be
            discarded.
        detect_ids (:obj:`List[int]`): **default = [0]**. |br|
            List of object class IDs to be detected. To detect all classes,
            refer to the :ref:`tech note <general-object-detection-ids>`.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.

    References:
        EfficientDet: Scalable and Efficient Object Detection:
        https://arxiv.org/abs/1911.09070

        Code adapted from https://github.com/xuannianz/EfficientDet.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = efficientdet_model.EfficientDetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Takes an image as input and returns bboxes of objects specified
        in config.
        """
        image = cv2.cvtColor(inputs["img"], cv2.COLOR_BGR2RGB)
        bboxes, labels, scores = self.model.predict(image)
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}
        return outputs



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Track states and STrack to store information for each tracked detection.

Modifications include:
- Make BaseTrack an ABC and use abstractmethod decorator for clarity
    - Removed update() method as it's not used in the current implementation of
        JDE.
- Renamed tlbr to xyxy for consistency with other model nodes.
- Removed new_id argument from re_activate() since it's never used
"""

from abc import ABC, abstractmethod
from collections import deque
from typing import Deque, List

import numpy as np
import torch

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.kalman_filter import (
    KalmanFilter,
)
from peekingduck.pipeline.utils.bbox.transforms import tlwh2xyah


class TrackState:  # pylint: disable=too-few-public-methods
    """Numbered states of Track.

    Attributes:
        NEW: The Track is newly created.
        TRACKED: The Track is actively tracked.
        LOST: The Track is not found among the detections and is considered
            "lost".
        REMOVED: The Track has been lost for longer than the threshold and is
            to be removed.
    """

    NEW = 0
    TRACKED = 1
    LOST = 2
    REMOVED = 3


class BaseTrack(ABC):
    """Base Tracking class."""

    _count = 0

    track_id = 0
    is_activated = False
    state = TrackState.NEW

    features: Deque[np.ndarray] = deque([])
    curr_feature = None
    start_frame = 0
    frame_id = 0
    time_since_update = 0

    @property
    def end_frame(self) -> int:
        """The last frame ID where this is actively tracked."""
        return self.frame_id

    def mark_lost(self) -> None:
        """Marks the Track as lost."""
        self.state = TrackState.LOST

    def mark_removed(self) -> None:
        """Marks the Track for removal."""
        self.state = TrackState.REMOVED

    @abstractmethod
    def activate(self, kalman_filter: KalmanFilter, frame_id: int) -> None:
        """Starts a new tracklet.

        Args:
            kalman_filter (KalmanFilter): Kalman filter for state estimation.
            frame_id (int): Current frame ID.
        """

    @abstractmethod
    def update(
        self, new_track: "STrack", frame_id: int, update_feature: bool = True
    ) -> None:
        """Updates a matched track.

        Args:
            new_track (STrack): New STrack.
            frame_id (int): Frame ID.
            update_feature (bool, optional): Update feature. Defaults to True.
        """

    @staticmethod
    def next_id() -> int:
        """The next track ID."""
        BaseTrack._count += 1
        return BaseTrack._count


class STrack(BaseTrack):  # pylint: disable=too-many-instance-attributes
    """Handles information of a single track.

    Args:
        tlwh (np.ndarray): Bounding box in (top left x, top left y, width,
            height) format.
        score (torch.Tensor): Detection confidence score.
        feat (np.ndarray): Embeddings.
        buffer_size (int): Maximum of past embeddings to store.
    """

    shared_kalman = KalmanFilter()

    def __init__(
        self,
        tlwh: np.ndarray,
        score: torch.Tensor,
        feat: np.ndarray,
        buffer_size: int = 30,
    ) -> None:
        self._tlwh = np.asarray(tlwh, dtype=np.float)
        self.score = score

        self.kalman_filter: KalmanFilter
        self.mean = None
        self.covariance = None

        self.is_activated = False
        self.tracklet_len = 0

        self.smooth_feat = None
        self.update_features(feat)
        self.features = deque([], maxlen=buffer_size)
        self.alpha = 0.9

    def __repr__(self) -> str:
        return f"OT_{self.track_id}_({self.start_frame}-{self.end_frame})"

    @property
    def tlwh(self) -> np.ndarray:
        """The current position in bounding box format `(top left x,
        top left y, width, height)`.
        """
        if self.mean is None:
            return self._tlwh.copy()
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    @property
    def xyah(self) -> np.ndarray:
        """The current position in bounding box to format `(center x, center y,
        aspect ratio, height)`, where the aspect ratio is `width / height`.
        """
        return tlwh2xyah(self.tlwh)

    @property
    def xyxy(self) -> np.ndarray:
        """The current position in bounding box format `(x1, y1, x2, y2)` where
        (x1, y1) is top left and (x2, y2) is bottom right.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    def activate(self, kalman_filter: KalmanFilter, frame_id: int) -> None:
        """Starts a new tracklet.

        Args:
            kalman_filter (KalmanFilter): Kalman filter for state estimation.
            frame_id (int): Current frame ID.
        """
        self.kalman_filter = kalman_filter
        self.track_id = self.next_id()
        self.mean, self.covariance = self.kalman_filter.initiate(tlwh2xyah(self._tlwh))

        self.tracklet_len = 0
        self.state = TrackState.TRACKED
        if frame_id == 1:
            self.is_activated = True
        self.frame_id = frame_id
        self.start_frame = frame_id

    def re_activate(self, new_track: "STrack", frame_id: int) -> None:
        """Re-activates STrack.

        Args:
            new_track (STrack): New STrack.
            frame_id (int): Current frame ID.
        """
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, tlwh2xyah(new_track.tlwh)
        )

        self.update_features(new_track.curr_feat)
        self.tracklet_len = 0
        self.state = TrackState.TRACKED
        self.is_activated = True
        self.frame_id = frame_id

    def update(
        self, new_track: "STrack", frame_id: int, update_feature: bool = True
    ) -> None:
        """Updates a matched track.

        Args:
            new_track (STrack): New STrack.
            frame_id (int): Frame ID.
            update_feature (bool, optional): Update feature. Defaults to True.
        """
        self.frame_id = frame_id
        self.tracklet_len += 1

        new_tlwh = new_track.tlwh
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, tlwh2xyah(new_tlwh)
        )
        self.state = TrackState.TRACKED
        self.is_activated = True

        self.score = new_track.score
        if update_feature:
            self.update_features(new_track.curr_feat)

    def update_features(self, feat: np.ndarray) -> None:
        """Updates the features (embeddings).

        Args:
            feat (np.ndarray): Embeddings.
        """
        feat /= np.linalg.norm(feat)
        self.curr_feat = feat
        if self.smooth_feat is None:
            self.smooth_feat = feat
        else:
            self.smooth_feat = self.alpha * self.smooth_feat + (1 - self.alpha) * feat
        self.features.append(feat)
        self.smooth_feat /= np.linalg.norm(self.smooth_feat)

    @staticmethod
    def multi_predict(stracks: List["STrack"]) -> None:
        """Runs the vectorized version of Kalman filter prediction step.

        Args:
            stracks (List[STrack]): List of STrack.
        """
        if not stracks:
            return
        multi_mean = np.asarray([st.mean.copy() for st in stracks])  # type: ignore
        multi_covariance = np.asarray([st.covariance for st in stracks])
        for i, strack in enumerate(stracks):
            if strack.state != TrackState.TRACKED:
                multi_mean[i][7] = 0
        multi_mean, multi_covariance = STrack.shared_kalman.multi_predict(
            multi_mean, multi_covariance
        )
        for i, (mean, cov) in enumerate(zip(multi_mean, multi_covariance)):
            stracks[i].mean = mean
            stracks[i].covariance = cov



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
functions for drawing legend related UI components
"""

from typing import Any, Dict, List, Union

import cv2
import numpy as np
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA

from peekingduck.pipeline.nodes.draw.utils.constants import (
    BLACK,
    FILLED,
    PRIMARY_PALETTE,
    PRIMARY_PALETTE_LENGTH,
    SMALL_FONTSCALE,
    THICK,
    WHITE,
)
from peekingduck.pipeline.nodes.draw.utils.general import get_image_size


class Legend:
    """Legend class that uses available info to draw legend box on frame"""

    def __init__(self) -> None:
        self.legend_left_x = 15

        self.frame = None
        self.legend_starting_y = 0
        self.delta_y = 0
        self.legend_height = 0

    def draw(
        self, inputs: Dict[str, Any], items: List[str], position: str
    ) -> np.ndarray:
        """Draw legends onto image

        Args:
            inputs (dict): dictionary of all available inputs for drawing the legend
            items (list): list of items to be drawn in legend
            position (str): used to control whether legend box is drawn on top or bottom
        """
        self.frame = inputs["img"]

        self.legend_height = self._get_legend_height(inputs, items)
        self._set_legend_variables(position)

        self._draw_legend_box(self.frame)
        y_pos = self.legend_starting_y + 20
        for item in items:
            if item == "zone_count":
                self._draw_zone_count(self.frame, y_pos, inputs[item])
            else:
                self.draw_item_info(self.frame, y_pos, item, inputs[item])
            y_pos += 20

    def draw_item_info(
        self,
        frame: np.ndarray,
        y_pos: int,
        item_name: str,
        item_info: Union[int, float, str],
    ) -> None:
        """Draw item name followed by item info onto frame. If item info is
        of float type, it will be displayed in 2 decimal places.

        Args:
            frame (np.array): image of current frame
            y_pos (int): y_position to draw the count text
            item_name (str): name of the legend item
            item_info: Union[int, float, str]: info contained by the legend item
        """
        if isinstance(item_info, (int, float, str)):
            pass
        else:
            raise TypeError(
                f"With the exception of the 'zone_count' data type, "
                f"the draw.legend node only draws values that are of type 'int', 'float' or 'str' "
                f"within the legend box. The value: {item_info} from the data type: {item_name} "
                f"is of type: {type(item_info)} and is unable to be drawn."
            )

        if isinstance(item_info, float):
            text = f"{item_name.upper()}: {item_info:.2f}"
        else:
            text = f"{item_name.upper()}: {str(item_info)}"
        cv2.putText(
            frame,
            text,
            (self.legend_left_x + 10, y_pos),
            FONT_HERSHEY_SIMPLEX,
            SMALL_FONTSCALE,
            WHITE,
            THICK,
            LINE_AA,
        )

    def _draw_zone_count(
        self, frame: np.ndarray, y_pos: int, counts: List[int]
    ) -> None:
        """Draw zone counts of all zones onto frame image

        Args:
            frame (np.array): image of current frame
            y_pos (int): y position to draw the count info text
            counts (list): list of zone counts
        """
        text = "-ZONE COUNTS-"
        cv2.putText(
            frame,
            text,
            (self.legend_left_x + 10, y_pos),
            FONT_HERSHEY_SIMPLEX,
            SMALL_FONTSCALE,
            WHITE,
            THICK,
            LINE_AA,
        )
        for i, count in enumerate(counts):
            y_pos += 20
            cv2.rectangle(
                frame,
                (self.legend_left_x + 10, y_pos - 14),
                (self.legend_left_x + 30, y_pos + 5),
                PRIMARY_PALETTE[(i + 1) % PRIMARY_PALETTE_LENGTH],
                FILLED,
            )
            text = f" ZONE-{i+1}: {count}"
            cv2.putText(
                frame,
                text,
                (40, y_pos),
                FONT_HERSHEY_SIMPLEX,
                SMALL_FONTSCALE,
                WHITE,
                THICK,
                LINE_AA,
            )

    def _draw_legend_box(self, frame: np.ndarray) -> None:
        """draw pts of selected object onto frame

        Args:
            frame (np.array): image of current frame
            zone_count (List[float]): object count, likely people, of each zone used
            in the zone analytics
        """
        assert self.legend_height is not None
        overlay = frame.copy()
        cv2.rectangle(
            overlay,
            (self.legend_left_x, self.legend_starting_y),
            (self.legend_left_x + 150, self.legend_starting_y + self.legend_height),
            BLACK,
            FILLED,
        )
        # apply the overlay
        cv2.addWeighted(overlay, 0.3, frame, 0.7, 0, frame)

    @staticmethod
    def _get_legend_height(inputs: Dict[str, Any], items: List[str]) -> int:
        """Get height of legend box needed to contain all items drawn"""
        no_of_items = len(items)
        if "zone_count" in items:
            # increase the number of items according to number of zones
            no_of_items += len(inputs["zone_count"])
        return 12 * no_of_items + 8 * (no_of_items - 1) + 20

    def _set_legend_variables(self, position: str) -> None:
        assert self.legend_height != 0
        if position == "top":
            self.legend_starting_y = 10
        else:
            _, image_height = get_image_size(self.frame)
            self.legend_starting_y = image_height - 10 - self.legend_height



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""👨‍👩‍👧‍👦 Congested Scene Recognition network: Dilated convolutional neural
networks for understanding the highly congested scenes.
"""

from typing import Any, Dict

import cv2

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.model.csrnetv1 import csrnet_model


class Node(AbstractNode):
    """Initializes and uses CSRNet model to predict the density map and crowd
    count.

    The csrnet node is capable of predicting the number of people in dense and
    sparse crowds. The dense and sparse crowd models were trained using data
    from ShanghaiTech Part A and ShanghaiTech Part B respectively. As the
    models were trained to recognize congested scenes, the estimates are less
    accurate if the number of people are low (e.g. less than 10).

    Inputs:
        |img_data|

    Outputs:
        |density_map_data|

        |count_data|

    Configs:
        model_type (:obj:`str`): **{"dense", "sparse"}, default="sparse"**. |br|
            Defines the type of CSRNet model to be used. The node uses the
            sparse crowd model by default and can be changed to using the dense
            crowd model. As a rule of thumb, the dense crowd model should be
            used if the people in a given image or video frame are packed
            shoulder to shoulder, e.g., stadiums.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        width (:obj:`int`): **default = 640**. |br|
            By default, the width of an image will be resized to 640 for
            inference. The height of the image will be resized proportionally
            to preserve its aspect ratio. In general, decreasing the width of
            an image will improve inference speed. However, this might impact
            the accuracy of the model.

    References:
        CSRNet: Dilated Convolutional Neural Networks for Understanding the
        Highly Congested Scenes:
        https://arxiv.org/pdf/1802.10062.pdf

        Model weights trained by https://github.com/Neerajj9/CSRNet-keras

        Inference code adapted from https://github.com/Neerajj9/CSRNet-keras
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = csrnet_model.CSRNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads in image frames and returns the density map and crowd count.

        Args:
            inputs (dict): Dictionary of inputs with key "img".

        Returns:
            outputs (dict): csrnet output in dictionary format with keys
            "density_map" and "count".
        """
        image = cv2.cvtColor(inputs["img"], cv2.COLOR_BGR2RGB)
        density_map, crowd_count = self.model.predict(image)
        outputs = {"density_map": density_map, "count": crowd_count}
        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of poses for HRNet
"""

import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.postprocessing import (
    affine_transform_xy,
    get_keypoint_conns,
    get_valid_keypoints,
    reshape_heatmaps,
    scale_transform,
)
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.preprocessing import (
    crop_and_resize,
    tlwh2xywh,
)
from peekingduck.pipeline.utils.bbox.transforms import xyxyn2tlwh
from peekingduck.utils.graph_functions import load_graph


class Detector:  # pylint: disable=too-few-public-methods
    """Detector class to handle detection of poses for HRNet."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        model_type: str,
        model_file: Dict[str, str],
        model_nodes: Dict[str, List[str]],
        resolution: Dict[str, int],
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]
        self.model_nodes = model_nodes
        self.resolution = resolution
        self.score_threshold = score_threshold

        self.hrnet = self._create_hrnet_model()

    def predict(
        self, frame: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """HRnet prediction function.

        Args:
            frame (np.ndarray): Image in numpy array.
            bboxes (np.ndarray): Array of detected bboxes.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple containing list of
            bboxes and pose related info, i.e., coordinates, scores, and
            connections
        """
        cropped_frames, affine_matrices, frame_size = self._preprocess(frame, bboxes)
        heatmaps = self.hrnet(tf.cast(cropped_frames, float))[0]

        cropped_frames_scale = [cropped_frames.shape[2], cropped_frames.shape[1]]
        poses, keypoint_scores, keypoint_conns = self._postprocess(
            heatmaps.numpy(), affine_matrices, cropped_frames_scale, frame_size
        )

        return poses, keypoint_scores, keypoint_conns

    def _create_hrnet_model(self) -> Callable:
        resolution_tuple = (self.resolution["height"], self.resolution["width"])
        self.logger.info(
            "HRNet graph model loaded with following configs:\n\t"
            f"Resolution: {resolution_tuple},\n\t"
            f"Score threshold: {self.score_threshold}"
        )
        return self._load_hrnet_weights()

    def _load_hrnet_weights(self) -> Callable:
        return load_graph(
            str(self.model_path),
            inputs=self.model_nodes["inputs"],
            outputs=self.model_nodes["outputs"],
        )

    def _postprocess(  # pylint: disable=too-many-locals
        self,
        heatmaps: np.ndarray,
        affine_matrices: np.ndarray,
        cropped_frames_scale: List[int],
        frame_size: Tuple[int, int],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Post processes output heatmaps to required keypoint arrays.

        Args:
            heatmaps (np.ndarray): Output heatmaps from hrnet network.
            affine_matrices (np.ndarray): transformation matrices of preprocess
                cropping.
            cropped_frames_scale (List[int]): Shape of cropped bboxes.
            frame_size (Tuple[int, int]): Size of original image.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Tuple containing array
            of bboxes and pose related info, i.e., coordinates, scores, and
            connections
        """
        batch, out_h, out_w, num_joints = heatmaps.shape
        heatmaps_reshaped = reshape_heatmaps(heatmaps)

        max_idxs = np.argmax(heatmaps_reshaped, 2)
        kp_scores = np.amax(heatmaps_reshaped, 2)
        keypoints = (
            np.repeat(max_idxs, 2).reshape(batch, num_joints, -1).astype(np.float32)
        )
        keypoints[:, :, 0] = (keypoints[:, :, 0]) % out_w
        keypoints[:, :, 1] = np.floor((keypoints[:, :, 1]) / out_w)

        keypoints = scale_transform(
            keypoints, in_scale=[out_w, out_h], out_scale=cropped_frames_scale
        )
        keypoints = affine_transform_xy(keypoints, affine_matrices)
        keypoints, kp_masks = get_valid_keypoints(
            keypoints, kp_scores, batch, self.score_threshold
        )

        normalized_keypoints = keypoints / frame_size
        keypoint_conns = get_keypoint_conns(normalized_keypoints, kp_masks)

        return normalized_keypoints, kp_scores, keypoint_conns

    def _preprocess(
        self, frame: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Tuple[int, int]]:
        """Crops bboxes while preserving aspect ratio.

        Args:
            frame (np.ndarray): Input image in numpy array.
            bboxes (np.ndarray): Array of detected bboxes.

        Returns:
            (Tuple[np.ndarray, np.ndarray, Tuple[int, int]]): Array of cropped
            bboxes, transformation matrices, and original frame size.
        """
        frame = frame / 255.0
        frame_size = (frame.shape[1], frame.shape[0])
        cropped_size = (self.resolution["width"], self.resolution["height"])

        tlwhs = xyxyn2tlwh(bboxes, frame.shape[0] - 1, frame.shape[1] - 1)
        xywhs = tlwh2xywh(tlwhs, self.resolution["width"] / self.resolution["height"])
        cropped_imgs, affine_matrices = crop_and_resize(frame, xywhs, cropped_size)

        return np.array(cropped_imgs), affine_matrices, frame_size



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preprocessing functions for input nodes
"""

import logging
from typing import Any, Tuple

import cv2
import numpy as np

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_res(stream: Any) -> Tuple[int, int]:
    """Gets the resolution for the video frame"""
    width = int(stream.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(stream.get(cv2.CAP_PROP_FRAME_HEIGHT))

    return width, height


def mirror(frame: np.ndarray) -> np.ndarray:
    """Mirrors a video frame."""
    return cv2.flip(frame, 1)


def resize_image(frame: np.ndarray, desired_width: int, desired_height: int) -> Any:
    """function that resizes the image input
    to the desired dimensions

    Args:
        frame (np.array): image
        desired_width: width of the resized image
        desired_height: height of the resized image

    Returns:
        image (np.array): returns a scaled image depending on the
        desired wight and height
    """
    return cv2.resize(frame, (desired_width, desired_height))



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CLI functions for PeekingDuck.
"""

import logging
import math
import tempfile
from pathlib import Path
from time import perf_counter
from typing import Any, Dict, List, Optional, Tuple, Union

import click
import yaml

from peekingduck import __version__
from peekingduck.declarative_loader import PEEKINGDUCK_NODE_TYPES, DeclarativeLoader
from peekingduck.runner import Runner
from peekingduck.utils.create_node_helper import (
    create_config_and_script_files,
    ensure_relative_path,
    ensure_valid_name,
    ensure_valid_name_partial,
    ensure_valid_type,
    ensure_valid_type_partial,
    get_config_and_script_paths,
    verify_option,
)
from peekingduck.utils.deprecation import deprecate
from peekingduck.utils.logger import LoggerSetup

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def create_custom_folder(custom_folder_name: str) -> None:
    """Makes custom nodes folder to create custom nodes.

    Args:
        custom_folder_name (:obj:`str`): Name of the custom nodes folder.
    """
    curr_dir = _get_cwd()
    custom_nodes_dir = curr_dir / "src" / custom_folder_name
    custom_nodes_config_dir = custom_nodes_dir / "configs"

    logger.info(f"Creating custom nodes folder in {custom_nodes_dir}")
    custom_nodes_dir.mkdir(parents=True, exist_ok=True)
    custom_nodes_config_dir.mkdir(parents=True, exist_ok=True)


def create_pipeline_config_yml(
    default_nodes: List[Union[str, Dict[str, Any]]] = None,
    default_path: Path = Path("pipeline_config.yml"),
) -> None:
    """Initializes the declarative *pipeline_config.yml*.

    Args:
        default_nodes (List[Union[str, Dict[str, Any]]]): A list of PeekingDuck
            node. For nodes with custom configuration, use a dictionary instead
            of a string.
        default_path (Path): Path of the pipeline config file.
    """
    # Default yml to be discussed
    if default_nodes is None:
        default_nodes = [
            {
                "input.visual": {
                    "source": "https://storage.googleapis.com/peekingduck/videos/wave.mp4"
                }
            },
            "model.posenet",
            "draw.poses",
            "output.screen",
        ]
    default_yml = dict(nodes=default_nodes)

    with open(default_path, "w") as yml_file:
        yaml.dump(default_yml, yml_file, default_flow_style=False)


@click.group(invoke_without_command=True)
@click.version_option(__version__)
@click.option("--verify_install", is_flag=True, help="Verify PeekingDuck installation")
@click.pass_context
def cli(ctx: click.Context, verify_install: bool) -> None:
    """
    PeekingDuck is a modular computer vision inference framework.

    Developed by Computer Vision Hub at AI Singapore.
    """
    if ctx.invoked_subcommand is None:
        if verify_install:
            _verify_install()
        else:
            print(ctx.get_help())


@cli.command()
@click.option(
    "--node_subdir",
    help=(
        "Path to the custom nodes directory, relative to the directory where "
        "the command is invoked."
    ),
    required=False,
)
@click.option(
    "--node_type",
    help=(
        "Node type, only accepts values from existing node types defined in "
        f"{PEEKINGDUCK_NODE_TYPES}."
    ),
    required=False,
)
@click.option(
    "--node_name",
    help=(
        "\b\nName of new custom node. The name cannot be a duplicate of an \n"
        "existing custom node. The name has the following requirements:\n"
        "- Minimum 2 characters\n"
        "- Can only contain alphanumeric characters, dashes and underscores \n"
        "  /[[a-zA-Z0-9_\\-]/\n"
        "- Must start with an alphabet\n"
        "- Must end with an alphanumeric character"
    ),
    required=False,
)
@click.option(
    "--config_path",
    help=(
        "List of nodes to parse. Will automatically create the config and script file "
        "for any custom nodes."
    ),
    required=False,
)
def create_node(
    node_subdir: Optional[str] = None,
    node_type: Optional[str] = None,
    node_name: Optional[str] = None,
    config_path: Optional[str] = None,
) -> None:
    """Automates the creation of a new custom node.

    If the options `node_subdir`, `node_type`, or `node_name` are not
    specified, users will be prompted them for the values while performing
    checks to ensure value validity.
    """
    project_dir = _get_cwd()
    node_type_choices = click.Choice(PEEKINGDUCK_NODE_TYPES)

    if config_path is not None and any(
        arg is not None for arg in (node_subdir, node_type, node_name)
    ):
        raise ValueError(
            "--config_path cannot be use with --node_subdir, --node_type, or "
            "--node_name!"
        )
    if config_path is not None:
        _create_nodes_from_config_file(config_path, project_dir, node_type_choices)
    else:
        click.secho("Creating new custom node...")
        node_subdir = verify_option(node_subdir, value_proc=ensure_relative_path)
        if node_subdir is None:
            node_subdir = click.prompt(
                f"Enter node directory relative to {project_dir}/src",
                default="custom_nodes",
                value_proc=ensure_relative_path,
            )
        node_dir = project_dir / "src" / node_subdir

        node_type = verify_option(
            node_type, value_proc=ensure_valid_type_partial(node_type_choices)
        )
        if node_type is None:
            node_type = click.prompt(
                f"Select node type ({', '.join(node_type_choices.choices)})",
                value_proc=ensure_valid_type_partial(node_type_choices),
            )

        node_name = verify_option(
            node_name, value_proc=ensure_valid_name_partial(node_dir, node_type)
        )
        if node_name is None:
            node_name = click.prompt(
                "Enter node name",
                default="my_custom_node",
                value_proc=ensure_valid_name_partial(node_dir, node_type),
            )

        created_paths = get_config_and_script_paths(
            node_dir, ("configs", node_type), node_type, node_name
        )
        click.echo(f"\nNode directory:\t{node_dir}")
        click.echo(f"Node type:\t{node_type}")
        click.echo(f"Node name:\t{node_name}")
        click.echo("\nCreating the following files:")
        click.echo(f"\tConfig file: {created_paths['config']}")
        click.echo(f"\tScript file: {created_paths['script']}")

        proceed = click.confirm("Proceed?", default=True)
        if proceed:
            create_config_and_script_files(created_paths)
            click.echo("Created node!")
        else:
            click.echo("Aborted!")


@cli.command()
@click.option("--custom_folder_name", default="custom_nodes")
def init(custom_folder_name: str) -> None:
    """Initialize a PeekingDuck project"""
    print("Welcome to PeekingDuck!")
    create_custom_folder(custom_folder_name)
    create_pipeline_config_yml()


@cli.command()
@click.argument("type_name", required=False)
def nodes(type_name: str = None) -> None:
    """Lists available nodes in PeekingDuck. When no argument is given, all
    available nodes will be listed. When the node type is given as an argument,
    all available nodes in the specified node type will be listed.

    Args:
        type_name (str): input, model, dabble, draw or output
    """
    config_dir = Path(__file__).resolve().parent / "configs"

    if type_name is None:
        node_types = PEEKINGDUCK_NODE_TYPES
    else:
        node_types = [type_name]

    for node_type in node_types:
        click.secho("\nPeekingDuck has the following ", bold=True, nl=False)
        click.secho(f"{node_type} ", fg="red", bold=True, nl=False)
        click.secho("nodes:", bold=True)

        node_names = [path.stem for path in (config_dir / node_type).glob("*.yml")]
        max_length = _len_enumerate(max(enumerate(node_names), key=_len_enumerate))
        for num, node_name in enumerate(node_names):
            idx = num + 1
            url = _get_node_url(node_type, node_name)
            node_name_width = max_length + 1 - _num_digits(idx)

            click.secho(f"{idx}:", nl=False)
            click.secho(f"{node_name: <{node_name_width}}", bold=True, nl=False)
            click.secho("Info: ", fg="yellow", nl=False)
            click.secho(url)

    click.secho("\n")


@cli.command()
@click.option(
    "--config_path",
    default=None,
    type=click.Path(),
    help=(
        "List of nodes to run. None assumes pipeline_config.yml at current working directory"
    ),
)
@click.option(
    "--log_level",
    default="info",
    help="""Modify log level {"critical", "error", "warning", "info", "debug"}""",
)
@click.option(
    "--node_config",
    default="None",
    help="""Modify node configs by wrapping desired configs in a JSON string.\n
        Example: --node_config '{"node_name": {"param_1": var_1}}'""",
)
@click.option(
    "--num_iter",
    default=None,
    type=int,
    help="Stop pipeline after running this number of iterations",
)
def run(
    config_path: str,
    log_level: str,
    node_config: str,
    num_iter: int,
    nodes_parent_dir: str = "src",
) -> None:
    """Runs PeekingDuck"""
    LoggerSetup.set_log_level(log_level)

    if config_path is None:
        curr_dir = _get_cwd()
        if (curr_dir / "pipeline_config.yml").is_file():
            config_path = curr_dir / "pipeline_config.yml"
        elif (curr_dir / "run_config.yml").is_file():
            deprecate(
                "using 'run_config.yml' as the default pipeline configuration "
                "file is deprecated and will be removed in the future. Please "
                "use 'pipeline_config.yml' instead.",
                2,
            )
            config_path = curr_dir / "run_config.yml"
        else:
            config_path = curr_dir / "pipeline_config.yml"
    pipeline_config_path = Path(config_path)

    start_time = perf_counter()
    runner = Runner(
        pipeline_path=pipeline_config_path,
        config_updates_cli=node_config,
        custom_nodes_parent_subdir=nodes_parent_dir,
        num_iter=num_iter,
    )
    end_time = perf_counter()
    logger.debug(f"Startup time = {end_time - start_time:.2f} sec")
    runner.run()


def _create_nodes_from_config_file(
    config_path: str, project_dir: Path, node_type_choices: click.Choice
) -> None:
    """Creates custom nodes declared in the pipeline config file."""
    pipeline_path = Path(config_path)
    if not pipeline_path.is_absolute():
        pipeline_path = (project_dir / pipeline_path).resolve()
    if not pipeline_path.exists():
        raise FileNotFoundError(
            f"Config file '{config_path}' is not found at {pipeline_path}!"
        )
    logger.info(f"Creating custom nodes declared in {pipeline_path}.")
    # Load run config with DeclarativeLoader to ensure consistency
    loader = DeclarativeLoader(pipeline_path, "None", "src")
    try:
        node_subdir = project_dir / "src" / loader.custom_nodes_dir
    except AttributeError as custom_nodes_no_exist:
        raise ValueError(
            f"Config file '{config_path}' does not contain custom nodes!"
        ) from custom_nodes_no_exist

    for node_str, _ in loader.node_list:
        try:
            node_subdir, node_type, node_name = node_str.split(".")
        except ValueError:
            continue
        try:
            # Check node string formatting
            ensure_relative_path(node_subdir)
            node_dir = project_dir / "src" / node_subdir
            ensure_valid_type(node_type_choices, node_type)
            ensure_valid_name(node_dir, node_type, node_name)
        except click.exceptions.UsageError as err:
            logger.warning(
                f"{node_str} contains invalid formatting: '{err.message}'. "
                "Skipping..."
            )
            continue
        created_paths = get_config_and_script_paths(
            node_dir, ("configs", node_type), node_type, node_name
        )
        logger.info(
            f"Creating files for {node_str}:\n\t"
            f"Config file: {created_paths['config']}\n\t"
            f"Script file: {created_paths['script']}"
        )
        create_config_and_script_files(created_paths)


def _get_cwd() -> Path:
    return Path.cwd()


def _get_node_url(node_type: str, node_name: str) -> str:
    """Constructs the URL to documentation of the specified node.

    Args:
        node_type (str): One of input, model, dabble, draw or output.
        node_name (str): Name of the node.

    Returns:
        (str): Full URL to the documentation of the specified node.
    """
    node_path = f"{node_type}.{node_name}"
    url_prefix = "https://peekingduck.readthedocs.io/en/stable/nodes/"
    url_postfix = ".html#module-"

    return f"{url_prefix}{node_path}{url_postfix}{node_path}"


def _len_enumerate(item: Tuple) -> int:
    """Calculates the string length of an enumerate item while accounting for
    the number of digits of the index.

    Args:
        item (Tuple): An item return by ``enumerate``, contains
            ``(index, element)`` where ``element`` is a ``str``.

    Returns:
        (int): Sum of length of ``element`` and the number of digits of its
            index.
    """
    return _num_digits(item[0] + 1) + len(item[1])


def _num_digits(number: int) -> int:
    """The number of digits of the given number.

    Args:
        number (int): The given number. It is assumed ``number > 0``.

    Returns:
        (int): Number of digits in the given number.
    """
    return int(math.log10(number))


def _verify_install() -> None:
    """Verifies PeekingDuck installation by running object detection on
    'wave.mp4'.
    """
    LoggerSetup.set_log_level("info")

    with tempfile.TemporaryDirectory() as tmp_dir:
        pipeline_config_path = Path(tmp_dir) / "verification_pipeline.yml"

        create_pipeline_config_yml(
            [
                {
                    "input.visual": {
                        "source": "https://storage.googleapis.com/peekingduck/videos/wave.mp4"
                    }
                },
                "model.yolo",
                "draw.bbox",
                "output.screen",
            ],
            pipeline_config_path,
        )

        print(tmp_dir)

        runner = Runner(
            pipeline_path=pipeline_config_path,
            config_updates_cli="None",
            custom_nodes_parent_subdir="src",
            num_iter=None,
        )
        runner.run()



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CSRNet model with model types: sparse and dense
"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.csrnetv1.csrnet_files.predictor import Predictor


class CSRNetModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """CSRNet model with model types: sparse and dense"""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds("width", "(0, +inf]")

        model_dir = self.download_weights()
        self.predictor = Predictor(
            model_dir,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["width"],
        )

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, int]:
        """Predicts density map and crowd count from frame.

        Args:
            frame (np.ndarray): input frame.

        Returns:
            density_map (np.ndarray): predicted density map.
            crowd_count (int): predicted count of people.
        """
        assert isinstance(frame, np.ndarray)

        # return density_map, crowd_count
        return self.predictor.predict_count_from_image(frame)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright 2021 Megvii, Base Detection
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Network blocks for constructing the YOLOX model.

Modifications include:
- BaseConv
    - Removed get_activations, uses SiLU only
    - Removed groups and bias arguments, uses group=1 and bias=False for
        nn.Conv2d only
- CSPLayer
    - Removed expansion, uses 0.5 only
- Remove SiLU export-friendly class
- Removed unused DWConv and ResLayer class
- Removed depthwise and act arguments
- Refactor and formatting
"""

from typing import Tuple

import torch
import torch.nn as nn


class BaseConv(nn.Module):
    """A Conv2d -> Batchnorm -> SiLU block."""

    # pylint: disable=invalid-name
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        ksize: int,
        stride: int,
    ) -> None:
        super().__init__()
        # same padding
        pad = (ksize - 1) // 2
        self.conv = nn.Conv2d(in_channels, out_channels, ksize, stride, pad, bias=False)
        self.bn = nn.BatchNorm2d(out_channels)
        self.act = nn.SiLU(inplace=True)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call."""
        return self.act(self.bn(self.conv(inputs)))

    def fuseforward(self, inputs: torch.Tensor) -> torch.Tensor:
        """The computation performed at every call when conv and batch norm
        layers are fused.
        """
        return self.act(self.conv(inputs))


class Bottleneck(nn.Module):
    """Standard bottleneck."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        shortcut: bool = True,
        expansion: float = 0.5,
    ) -> None:
        super().__init__()
        hidden_channels = int(out_channels * expansion)
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, 1)
        self.conv2 = BaseConv(hidden_channels, out_channels, 3, 1)
        self.use_add = shortcut and in_channels == out_channels

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call."""
        outputs = self.conv2(self.conv1(inputs))
        if self.use_add:
            outputs = outputs + inputs
        return outputs


class SPPBottleneck(nn.Module):
    """Spatial pyramid pooling layer used in YOLOv3-SPP."""

    # pylint: disable=invalid-name
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_sizes: Tuple[int, int, int] = (5, 9, 13),
    ) -> None:
        super().__init__()
        hidden_channels = in_channels // 2
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(ks, 1, ks // 2) for ks in kernel_sizes])
        conv2_channels = hidden_channels * (len(kernel_sizes) + 1)
        self.conv2 = BaseConv(conv2_channels, out_channels, 1, 1)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call."""
        inputs = self.conv1(inputs)
        inputs = torch.cat([inputs] + [m(inputs) for m in self.m], dim=1)
        inputs = self.conv2(inputs)
        return inputs


class CSPLayer(nn.Module):
    """C3 in yolov5, CSP Bottleneck with 3 convolutions."""

    # pylint: disable=invalid-name
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_blocks: int = 1,
        shortcut: bool = True,
    ) -> None:
        """
        Args:
            in_channels (int): Input channels.
            out_channels (int): Output channels.
            num_blocks (int): Number of Bottlenecks, default = 1.
        """
        # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        hidden_channels = int(out_channels * 0.5)  # hidden channels
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, 1)
        self.conv2 = BaseConv(in_channels, hidden_channels, 1, 1)
        self.conv3 = BaseConv(2 * hidden_channels, out_channels, 1, 1)
        self.m = nn.Sequential(
            *[
                Bottleneck(hidden_channels, hidden_channels, shortcut, 1.0)
                for _ in range(num_blocks)
            ]
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call."""
        inputs_1 = self.conv1(inputs)
        inputs_2 = self.conv2(inputs)
        inputs_1 = self.m(inputs_1)
        inputs = torch.cat((inputs_1, inputs_2), dim=1)
        return self.conv3(inputs)


class Focus(nn.Module):
    """Focus width and height information into channel space."""

    # pylint: disable=invalid-name
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        ksize: int = 1,
        stride: int = 1,
    ) -> None:
        super().__init__()
        self.conv = BaseConv(in_channels * 4, out_channels, ksize, stride)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call."""
        # shape of x (b,c,w,h) -> y(b,4c,w/2,h/2)
        patch_top_left = inputs[..., ::2, ::2]
        patch_top_right = inputs[..., ::2, 1::2]
        patch_bot_left = inputs[..., 1::2, ::2]
        patch_bot_right = inputs[..., 1::2, 1::2]
        inputs = torch.cat(
            (patch_top_left, patch_bot_left, patch_top_right, patch_bot_right), dim=1
        )
        return self.conv(inputs)
